# -*- coding: utf-8 -*-
"""
Cool tools
~~~~~~~~~~

The tools for your .cool's.

:author: Cooltools developers
:license: MIT

"""
import logging

__version__ = "0.4.0rc1"

from . import io
from . import lib
from .lib import numutils, download_data, print_available_datasets, get_data_dir, download_file, get_md5sum



### Adaptation of Max Imakaev's fast txt matrix writer.

cimport cython
import os
import subprocess

from libc.stdlib cimport malloc, free
from libc.string cimport strcpy, strlen

import numpy as np
cimport numpy as np

cdef extern from "stdio.h":
    int sprintf(char *str, char *format, ...)

def commandExists(command):
    """
    Checks if the bash command exists.
    """
    command = command.split()[0]
    if subprocess.call(['which', command]) != 0:
        return False
    return True

def gzipWriter(filepath):
    """
    Creates a writing process with gzip or parallel gzip (pigz) attached to it.
    """
    filepath = os.path.abspath(filepath)
    with open(filepath, 'wb') as outFile:
        if commandExists("pigz"):
            writer = ["pigz", "-c", "-9"]
        else:
            writer = ["gzip", "-c", "-2"]

        pwrite = subprocess.Popen(
            writer,
            stdin=subprocess.PIPE,
            stdout=outFile,
            shell=False,
            bufsize=-1)
    return pwrite

def empty_func():
    return None

@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.wraparound(False)

def array2txt(
    mat,
    out,
    format_string=b'%.4lf',
    sep=b'\t',
    newline=b'\n',
    header=None,
    row_headers=None,
    max_element_len=100):
    """
    Dump a 2d array into a text file, optionally gzipped.
    This implementation if faster than the np.savetxt and it allows the user
    to provide column/row headers.

    Parameters
    ----------
    mat : a 2D numpy array of a list of lists of numbers (float/integer)

    out : str or file object
        Either:
        -- a path to the output file. If ends with .gz the output is gzipped
        -- a file object
        -- a stdin of a Popen object
        TIP: when using files/stdin do not forget to flush()/communicate().

    format_string : bytes, optional
        A printf-style formatting string to specify the coversion of
        the elements of the matrix into strings.

    sep : bytes, optional
        The column separator.

    newline : bytes, optional
        The newline separator.

    header : bytes, optional
        A header to prepend to the output file, is separated from the main table
        by a `newline`.

    row_headers : a list of bytes, optional
        Row headers to prepend to the output file, one per each row in `mat`.

    max_element_len : int
            The maximal length of the string representation of a matrix element,
        produced by sprintf(`format_string`). Used to preallocate memory.
    """


    cdef int N = len(mat)
    cdef int M = len(mat[0])

    if issubclass(type(out), str) or issubclass(type(out), bytearray):
        if out.endswith('.gz'):
            writer = gzipWriter(out)
            out_pipe = writer.stdin
            close_out_func = writer.communicate
        else:
            writer = open(out, 'wb')
            out_pipe = writer
            close_out_func = writer.flush
    elif hasattr(out, 'write'):
        out_pipe = out
        close_out_func = empty_func
    else:
        raise Exception('`out` must be either a file path or a file handle/stream')

    cdef np.ndarray[np.double_t, ndim=2] mat_ndarray = np.array(mat, dtype=np.double, order="C")

    cdef char* newline_cstr = newline
    cdef char* sep_cstr = sep
    cdef char* next_row_header
    cdef char* s_start
    cdef char* s_cur

    cdef int max_header_len = 0
    if row_headers is not None:
        max_header_len = max([len(row_header) for row_header in row_headers])

    s_start = <char *>malloc((max_element_len * M  + max_header_len) * sizeof(char))

    cdef double element
    cdef char* curStringTemplate
    template = b''.join([format_string, sep])
    curStringTemplate = template

    if header is not None:
        out_pipe.write(header)
        out_pipe.write(newline_cstr)

    cdef int i,j
    for i in xrange(N):
        s_cur = s_start
        if row_headers is not None:

            next_row_header = row_headers[i]
            s_cur = strcpy(s_cur, next_row_header)
            s_cur += sizeof(char) * strlen(next_row_header)

            s_cur = strcpy(s_cur, sep_cstr)
            s_cur += sizeof(char) * strlen(sep_cstr)

        for j in xrange(M):
            element = mat_ndarray[i,j]
            s_cur = s_cur + sprintf(s_cur, curStringTemplate, element)

        s_cur = strcpy(s_cur, newline_cstr)
        s_cur += sizeof(char) * strlen(newline_cstr)

        out_pipe.write(s_start)
    free(s_start)

    close_out_func()



import numpy as np
import pandas as pd

# Test data download requirements:
import requests
import os
import hashlib

URL_DATA = "https://raw.githubusercontent.com/open2c/cooltools/master/datasets/external_test_files.tsv"


def assign_supports(features, supports, labels=False, suffix=""):
    """
    Assign support regions to a table of genomic intervals.

    Parameters
    ----------
    features : DataFrame
        Dataframe with columns `chrom`, `start`, `end`
        or `chrom1`, `start1`, `end1`, `chrom2`, `start2`, `end2`
    supports : array-like
        Support areas

    """
    features = features.copy()
    supp_col = pd.Series(index=features.index, data=np.nan)

    c = "chrom" + suffix
    s = "start" + suffix
    e = "end" + suffix
    for col in (c, s, e):
        if col not in features.columns:
            raise ValueError(
                'Column "{}" not found in features data frame.'.format(col)
            )

    for i, region in enumerate(supports):
        # single-region support
        if len(region) in [3, 4]:
            sel = (features[c] == region[0]) & (features[e] > region[1])
            if region[2] is not None:
                sel &= features[s] < region[2]
        # paired-region support
        elif len(region) == 2:
            region1, region2 = region
            sel1 = (features[c] == region1[0]) & (features[e] > region1[1])
            if region1[2] is not None:
                sel1 &= features[s] < region1[2]
            sel2 = (features[c] == region2[0]) & (features[e] > region2[1])
            if region2[2] is not None:
                sel2 &= features[s] < region2[2]
            sel = sel1 | sel2
        supp_col.loc[sel] = i

    if labels:
        supp_col = supp_col.map(lambda i: supports[int(i)], na_action="ignore")

    return supp_col


def assign_regions_to_bins(bin_ids, regions_span):
    regions_binsorted = (
        regions_span[(regions_span["bin_start"] >= 0) & (regions_span["bin_end"] >= 0)]
            .sort_values(["bin_start", "bin_end"])
            .reset_index()
    )

    bin_reg_idx_lo = regions_span["bin_start"].searchsorted(bin_ids, "right") - 1
    bin_reg_idx_hi = regions_span["bin_end"].searchsorted(bin_ids, "right")
    mask_assigned = (bin_reg_idx_lo == bin_reg_idx_hi) & (bin_reg_idx_lo >= 0)

    region_ids = pd.array([pd.NA] * len(bin_ids))
    region_ids[mask_assigned] = regions_span["name"][bin_reg_idx_lo[mask_assigned]]

    return region_ids


def download_file(url, local_filename=None):
    """
    Download single file and return its name.

    """

    if local_filename is None:
        local_filename = url.split('/')[-1]
    print('downloading:', url, 'as', local_filename)
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return local_filename


def download_data(name="all",
                  cache=True,
                  data_dir=None,
                  ignore_checksum=False,
                  checksum_chunk=8192,
                  _url_info=URL_DATA):
    """
    Download the specified dataset or all available datasets ("all" option).
    Check available datasets with cooltools.print_available_datasets()

    By default, checksums are verified for downloaded and cached files.

    Parameters
    ----------
    name : str
        Name of the dataset. Default: "all"

    cache : boolean, optional
        If True, try to load from the local cache first, and save to the cache
        if a download is required. Default: True

    data_dir : string, optional
        The directory where to cache data; default is defined by :func:`get_data_dir`.

    ignore_checksum : boolean, optional
        Ignore checksum of the file (not recommended). Default: False

    checksum_chunk : int, optional
        Sets the chunksize for calculation of md5sum of the file.
        Used only if ignore_checksum is False.

    Returns
    -------
    Path to the file if a single file was downloaded, or a list (for 'all' download).

    """

    available_datasets = _get_datasets_info(_url_info)
    available_keys = list(map(lambda x: x['key'], available_datasets))

    if name in available_keys:
        keys = [name]
    elif name == "all":
        keys = available_keys
    else:
        raise Exception(
            """Dataset {key} is not available. 
            Available datasets: {datasets}
            Use print_available_datasets() to see the details. 
            """.format(key=name, datasets=','.join(available_keys)))

    data_dir = get_data_dir(data_dir)

    assert len(keys) > 0  # Checking that test data file is parsed successfully
    downloaded = []
    for data in available_datasets:

        key, url, local_filename, original_checksum = data["key"], data["link"], data["filename"], data["checksum"]

        if key not in keys:
            continue

        file_path = os.path.join(data_dir, local_filename)

        if cache and os.path.exists(file_path):
            if not ignore_checksum:
                checksum = get_md5sum(file_path, chunksize=checksum_chunk)
                assert checksum==original_checksum, \
f"""The downloaded {key} file in {local_filename} differs from original test: {url}
Re-reun with cache=False """
            downloaded += [file_path]
            continue

        elif cache:
            print("Test dataset {} (file {}) is not in the cache directory {}".format(key, local_filename, data_dir))

        file_path = download_file(url, file_path)
        if not ignore_checksum:
            checksum = get_md5sum(file_path, chunksize=checksum_chunk)
            assert checksum == original_checksum, \
                f"Download of {key} to {local_filename} failed. File differs from original test: {url}"

        downloaded += [file_path]

    # No files, return empty string
    if len(downloaded) == 0:
        return ''
    # A single file downloaded, return its name
    elif len(downloaded) == 1:
        return downloaded[0]
    # Multiple files downloaded, return list
    else:
        return downloaded


def get_md5sum(file_path, chunksize=8192):
    """
    Load file by chunks and return md5sum of a file.
    File might be rather large, and we don't load it into memory as a whole.
    Name is excluded from calculation of hash.

    Parameters
    ----------
    file_path : str
        Location of the file

    chunksize : int
        Size of the chunk

    Returns
    -------
    String with md5sum hex hash

    """

    md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        while True:
            file_fragment = f.read(chunksize)
            if not file_fragment:
                break
            md5.update(file_fragment)
        return md5.hexdigest()


def get_data_dir(data_dir=None):
    """
    Return a path to cache directory for example datasets.

    This directory is then used by :func:`download_data`.

    By default, this function uses the location of cooltools as the home folder
    and stores the files under ./datasets/ path there.

    Parameters
    ----------
    data_dir : str, optional
        Location of the home for test data storage

    Returns
    -------
    String with path to the folder with test data

    Notes
    -------
    Designed following seaborn download test data examples:
    https://github.com/mwaskom/seaborn/blob/4dd57d63e23e4628d8b7f41cdefdf64f7fefcf74/seaborn/utils.py#L427

    """

    if data_dir is None:
        data_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..', 'datasets/'))

    data_dir = os.path.expanduser(data_dir)
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    return os.path.join(data_dir, '')


def _get_datasets_info(url=URL_DATA):
    """
    Reports all available datasets in URL_DATA

    Requires internet connection if url is a remote link.

    Returns
    -------

    A list of datasets, where each element is a dictionary with key, url, local_filename and checksum fields.

    """

    # Info file is remote
    if url.startswith("https://") or url.startswith("http://"):
        datasets_metadata = requests.get(url, stream=True).iter_lines()
    # Info file is local
    else:
        datasets_metadata = open(url, 'rb').readlines()

    header = []
    datasets_parsed = []
    for i, line_metadata in enumerate(datasets_metadata):
        # Read header:
        if line_metadata.decode("utf-8")[0] == "#":
            header = line_metadata.decode("utf-8")[2:].strip().split("\t")
        # Read the rest of the file:
        else:
            if len(header)==0:
                raise Exception(f"Data info file {url} is corrupted, no header denoted by '#'.")
            data_line = line_metadata.decode("utf-8").strip().split("\t")
            data = dict(zip(header, data_line))
            data.update({'index': i})
            datasets_parsed.append(dict(data))

    return datasets_parsed


def print_available_datasets(url=URL_DATA):
    """
    Prints all available test datasets in URL_DATA

    Requires internet connection if url is a remote link.

    """

    datasets_parsed = _get_datasets_info(url)
    for data in datasets_parsed:
        print("""{index}) {key} : {comment} 
\tDownloaded from {link} 
\tStored as {filename} 
\tOriginal md5sum: {checksum}
""".format(**data))



import os.path as op
import subprocess
import sys

import numpy as np
import pandas as pd
from click.testing import CliRunner
from cooltools.cli import cli


def test_compartment_cli(request, tmpdir):
    in_cool = op.join(request.fspath.dirname, "data/sin_eigs_mat.cool")
    out_eig_prefix = op.join(tmpdir, "test.eigs")
    runner = CliRunner()
    result = runner.invoke(
        cli, [
            'call-compartments',
            '-o', out_eig_prefix,
            in_cool,
        ]
    )
    assert result.exit_code == 0
    test_eigs = pd.read_table(out_eig_prefix + ".cis.vecs.tsv", sep="\t")
    gb = test_eigs.groupby("chrom")
    for chrom in gb.groups:
        chrom_eigs = gb.get_group(chrom)
        r = np.abs(
            np.corrcoef(
                chrom_eigs.E1.values, np.sin(chrom_eigs.start * 2 * np.pi / 500)
            )[0, 1]
        )
        assert r > 0.95


def test_saddle_cli(request, tmpdir):
    in_cool = op.join(request.fspath.dirname, "data/sin_eigs_mat.cool")
    out_eig_prefix = op.join(tmpdir, "test.eigs")
    out_expected = op.join(tmpdir, "test.expected")
    out_saddle_prefix = op.join(tmpdir, "test.saddle")

    runner = CliRunner()
    result = runner.invoke(
        cli, [
            'call-compartments',
            '-o', out_eig_prefix,
            in_cool
        ]
    )
    assert result.exit_code == 0

    try:
        subprocess.check_output(
            f"python -m cooltools compute-expected {in_cool} > {out_expected}",
            shell=True,
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    runner = CliRunner()
    result = runner.invoke(
        cli, [
            "compute-saddle",
            "-o", out_saddle_prefix,
            "--range", "-0.5", "0.5",
            "--n-bins", "30",
            "--scale", "log",
            in_cool,
            f"{out_eig_prefix}.cis.vecs.tsv",
            out_expected
        ]
    )
    assert result.exit_code == 0

    log2_sad = np.log2(np.load(out_saddle_prefix + ".saddledump.npz")["saddledata"])
    bins = np.load(out_saddle_prefix + ".saddledump.npz")["binedges"]
    binmids = (bins[:-1] + bins[1:]) / 2
    log2_theor_sad = np.log2(1 + binmids[None, :] * binmids[:, None])

    log2_sad_flat = log2_sad[1:-1, 1:-1].flatten()
    log2_theor_sad_flat = log2_theor_sad.flatten()

    mask = np.isfinite(log2_sad_flat) & np.isfinite(log2_theor_sad_flat)

    cc = np.abs(np.corrcoef(log2_sad_flat[mask], log2_theor_sad_flat[mask])[0][1])

    assert cc > 0.9


def test_trans_compartment_cli(request, tmpdir):
    # somehow - it is E3 that captures sin-like plaid
    # pattern, instead of E1 - we'll keep it like that for now:
    in_cool = op.join(request.fspath.dirname, "data/sin_eigs_mat.cool")
    out_eig_prefix = op.join(tmpdir, "test.eigs")
    runner = CliRunner()
    result = runner.invoke(
        cli, [
            'call-compartments',
            '--contact-type', "trans",
            '-o', out_eig_prefix,
            in_cool,
        ]
    )
    assert result.exit_code == 0
    test_trans_eigs = pd.read_table(out_eig_prefix + ".trans.vecs.tsv", sep="\t")
    r = np.abs(
        np.corrcoef(
            test_trans_eigs.E1.values, np.sin(test_trans_eigs.start * 2 * np.pi / 500)
        )[0, 1]
    )
    assert r > 0.95


def test_trans_saddle_cli(request, tmpdir):
    in_cool = op.join(request.fspath.dirname, "data/sin_eigs_mat.cool")
    out_eig_prefix = op.join(tmpdir, "test.eigs")
    out_expected = op.join(tmpdir, "test.trans.expected")
    out_saddle_prefix = op.join(tmpdir, "test.trans.saddle")

    runner = CliRunner()
    result = runner.invoke(
        cli, [
            'call-compartments',
            '--contact-type', "trans",
            '-o', out_eig_prefix,
            in_cool
        ]
    )
    assert result.exit_code == 0

    try:
        subprocess.check_output(
            f"python -m cooltools compute-expected --contact-type trans {in_cool} > {out_expected}",
            shell=True,
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    runner = CliRunner()
    result = runner.invoke(
        cli, [
            "compute-saddle",
            "-o", out_saddle_prefix,
            "--contact-type", "trans",
            "--range", "-0.5", "0.5",
            "--n-bins", "30",
            "--scale", "log",
            in_cool,
            f"{out_eig_prefix}.trans.vecs.tsv",
            out_expected
        ]
    )
    assert result.exit_code == 0

    log2_sad = np.log2(np.load(out_saddle_prefix + ".saddledump.npz")["saddledata"])
    bins = np.load(out_saddle_prefix + ".saddledump.npz")["binedges"]
    binmids = (bins[:-1] + bins[1:]) / 2
    log2_theor_sad = np.log2(1 + binmids[None, :] * binmids[:, None])

    log2_sad_flat = log2_sad[1:-1, 1:-1].flatten()
    log2_theor_sad_flat = log2_theor_sad.flatten()

    mask = np.isfinite(log2_sad_flat) & np.isfinite(log2_theor_sad_flat)

    cc = np.abs(np.corrcoef(log2_sad_flat[mask], log2_theor_sad_flat[mask])[0][1])

    assert cc > 0.9


# def test_digitize_track(request):
#     pass


# def test_make_saddle(request):
#     pass


# def test_saddleplot(request):
#     pass


# def test_saddlestrength(request):
#     pass



import subprocess
import pandas as pd
import numpy as np
import h5py


# make chromsizes
with open("./test.chrom.sizes", "w") as chromsizes:
    chromsizes.write("chr1\t1000\n")
    chromsizes.write("chr2\t2000\n")
    chromsizes.write("chr3\t3000")

BIN_SIZE = 10
# make bins
subprocess.check_output(
    f"cooltools genome binnify ./test.chrom.sizes {BIN_SIZE} > ./test.10.bins",
    shell=True,
)

# make Hi-C data
bins = pd.read_table("./test.10.bins", sep="\t")
EIG_PERIOD_BP = 500
EIG_AMPLITUDE = np.sqrt(0.5)
SCALING = -2
MAX_CIS_COUNTS = 1e8
MAX_TRANS_COUNTS = 1e5

bins["eig"] = EIG_AMPLITUDE * np.sin(bins.start * 2 * np.pi / EIG_PERIOD_BP)
bins["key"] = 0
pixels = pd.merge(bins, bins, on="key", how="outer", suffixes=("1", "2"))
pixels.drop("key", axis="columns", inplace=True)
pixels["count"] = np.nan

cis = pixels.chrom1 == pixels.chrom2
pixels.loc[cis, "count"] = pixels[cis].eval(
    "@MAX_CIS_COUNTS * ((abs(start1-start2)+@BIN_SIZE)**@SCALING) * (1.0+eig1*eig2)"
)
pixels.loc[~cis, "count"] = pixels[~cis].eval("@MAX_TRANS_COUNTS * (1.0+eig1*eig2)")

pixels["count"] = pixels["count"].astype(int)
pixels[["chrom1", "start1", "end1", "chrom2", "start2", "end2", "count"]].to_csv(
    "./sin_eigs_mat.bg2.gz", sep="\t", index=False, header=None, compression="gzip"
)


# make a cooler
subprocess.check_output(
    "cooler load -f bg2 --count-as-float --tril-action drop "
    + f"./test.chrom.sizes:{BIN_SIZE} ./sin_eigs_mat.bg2.gz "
    + "./sin_eigs_mat.cool",
    shell=True,
)

# fake IC
f = h5py.File("./sin_eigs_mat.cool")
f["bins/weight"] = np.ones_like(f["bins/start"], dtype=float)
f["bins/weight"].attrs["ignore_diags"] = 2
f.close()



# inspired by:
# saddles.py by @nvictus
# https://github.com/nandankita/labUtilityTools
from functools import partial
import os.path as op
import sys
import pandas as pd
import numpy as np
import cooler
from bioframe import parse_regions
from .. import saddle

import click
from .util import validate_csv
from . import util
from . import cli


@cli.command()
@click.argument(
    "cool_path",
    metavar="COOL_PATH",
    type=str
)  # click.Path(exists=True, dir_okay=False),)
@click.argument(
    "track_path",
    metavar="TRACK_PATH",
    type=str,
    callback=partial(validate_csv, default_column="E1"),
)
@click.argument(
    "expected_path",
    metavar="EXPECTED_PATH",
    type=str,
    callback=partial(validate_csv, default_column="balanced.avg"),
)
@click.option(
    "-t", "--contact-type",
    help="Type of the contacts to aggregate",
    type=click.Choice(["cis", "trans"]),
    default="cis",
    show_default=True,
)
@click.option(
    "--min-dist",
    help="Minimal distance between bins to consider, bp. If negative, removes"
    "the first two diagonals of the data. Ignored with --contact-type trans.",
    type=int,
    default=-1,
    show_default=True,
)
@click.option(
    "--max-dist",
    help="Maximal distance between bins to consider, bp. Ignored, if negative."
    " Ignored with --contact-type trans.",
    type=int,
    default=-1,
    show_default=True,
)
@click.option(
    "-n", "--n-bins",
    help="Number of bins for digitizing track values.",
    type=int,
    default=50,
    show_default=True,
)
@click.option(
    "--quantiles",
    help="Bin the signal track into quantiles rather than by value.",
    is_flag=True,
    default=False,
)
@click.option(
    "--range",
    "range_",
    help="Low and high values used for binning genome-wide track values, e.g. "
    "if `range`=(-0.05, 0.05), `n-bins` equidistant bins would be generated. "
    "Use to prevent the extreme track values from exploding the bin range and "
    "to ensure consistent bins across several runs of `compute_saddle` command "
    "using different track files.",
    nargs=2,
    type=float,
)
@click.option(
    "--qrange",
    help="The fraction of the genome-wide range of the track values used to "
    "generate bins. E.g., if `qrange`=(0.02, 0.98) the lower bin would "
    "start at the 2nd percentile and the upper bin would end at the 98th "
    "percentile of the genome-wide signal. "
    "Use to prevent the extreme track values from exploding the bin range.",
    type=(float, float),
    default=(0.0, 1.0),
    show_default=True,
)
@click.option(
    "--weight-name",
    help="Use balancing weight with this name.",
    type=str,
    default="weight",
    show_default=True,
)
@click.option(
    "--strength/--no-strength",
    help="Compute and save compartment 'saddle strength' profile",
    is_flag=True,
    default=False,
)
@click.option(
    "--regions",
    help="Path to a BED file containing genomic regions "
    "for which saddleplot will be calculated. Region names can "
    "be provided in a 4th column and should match regions and "
    "their names in expected.",
    type=click.Path(exists=True),
    required=False,
)
@click.option(
    "-o", "--out-prefix",
    help="Dump 'saddledata', 'binedges' and 'hist' arrays in a numpy-specific "
    ".npz container. Use numpy.load to load these arrays into a "
    "dict-like object. The digitized signal values are saved to a "
    "bedGraph-style TSV.",
    required=True,
)
@click.option(
    "--fig",
    type=click.Choice(["png", "jpg", "svg", "pdf", "ps", "eps"]),
    multiple=True,
    help="Generate a figure and save to a file of the specified format. "
    "If not specified - no image is generated. Repeat for multiple "
    "output formats.",
)
@click.option(
    "--scale",
    help="Value scale for the heatmap",
    type=click.Choice(["linear", "log"]),
    default="log",
    show_default=True,
)
@click.option(
    "--cmap",
    help="Name of matplotlib colormap",
    default="coolwarm",
    show_default=True
)
@click.option(
    "--vmin",
    help="Low value of the saddleplot colorbar. "
    "Note: value in original units irrespective of used scale, "
    "and therefore should be positive for both vmin and vmax.",
    type=float,
    default=0.5,
)
@click.option(
    "--vmax",
    help="High value of the saddleplot colorbar",
    type=float,
    default=2
)
@click.option(
    "--hist-color",
    help="Face color of histogram bar chart"
)
@click.option(
    "-v", "--verbose",
    help="Enable verbose output",
    is_flag=True,
    default=False
)
def compute_saddle(
    cool_path,
    track_path,
    expected_path,
    contact_type,
    min_dist,
    max_dist,
    n_bins,
    quantiles,
    range_,
    qrange,
    weight_name,
    strength,
    regions,
    out_prefix,
    fig,
    scale,
    cmap,
    vmin,
    vmax,
    hist_color,
    verbose,
):
    """
    Calculate saddle statistics and generate saddle plots for an arbitrary
    signal track on the genomic bins of a contact matrix.

    COOL_PATH : The paths to a .cool file with a balanced Hi-C map. Use the
    '::' syntax to specify a group path in a multicooler file.

    TRACK_PATH : The path to bedGraph-like file with a binned compartment track
    (eigenvector), including a header. Use the '::' syntax to specify a column
    name.

    EXPECTED_PATH : The paths to a tsv-like file with expected signal,
    including a header. Use the '::' syntax to specify a column name.

    Analysis will be performed for chromosomes referred to in TRACK_PATH, and
    therefore these chromosomes must be a subset of chromosomes referred to in
    COOL_PATH and EXPECTED_PATH.

    COOL_PATH, TRACK_PATH and EXPECTED_PATH must be binned at the same
    resolution (expect for  EXPECTED_PATH in case of trans contact type).

    EXPECTED_PATH must contain at least the following columns for cis contacts:
    'chrom', 'diag', 'n_valid', value_name and the following columns for trans
    contacts: 'chrom1', 'chrom2', 'n_valid', value_name value_name is controlled
    using options. Header must be present in a file.

    """
    clr = cooler.Cooler(cool_path)
    expected_path, expected_name = expected_path
    track_path, track_name = track_path

    if regions is None:
        regions = [(chrom, 0, clr.chromsizes[chrom]) for chrom in clr.chromnames]
        regions = parse_regions(regions)
        regions["name"] = clr.chromnames
    else:
        regions_buf, names = util.sniff_for_header(regions)
        regions = pd.read_csv(regions_buf, sep="\t", header=None)
        if regions.shape[1] not in (3, 4):
            raise ValueError(
                "The region file does not have three or four tab-delimited columns."
                "We expect a bed file with columns chrom, start, end, and optional name"
            )
        if regions.shape[1] == 4:
            regions = regions.rename(columns={0:"chrom",1:"start",2:"end",3:"name"})
            regions = parse_regions(regions)
        else:
            regions = regions.rename(columns={0:"chrom",1:"start",2:"end"})
            regions["name"] = list(regions.apply(lambda x: "{}:{}-{}".format(*x), axis=1))
            regions = parse_regions(regions)


    if vmin <= 0 or vmax <= 0:
        raise ValueError(
            "vmin and vmax values are in original units irrespective "
            "of used scale, and therefore should be positive"
        )

    # read expected and make preparations for validation,
    # it's contact_type dependent:
    if contact_type == "cis":
        # that's what we expect as column names:
        expected_columns = ["region", "diag", "n_valid", expected_name]
        # what would become a MultiIndex:
        expected_index = ["region", "diag"]
        # expected dtype as a rudimentary form of validation:
        expected_dtype = {
            "region": np.str,
            "diag": np.int64,
            "n_valid": np.int64,
            expected_name: np.float64,
        }
        # # unique list of chroms mentioned in expected_path:
        # get_exp_chroms = lambda df: df.index.get_level_values("region").unique()
        # # compute # of bins by comparing matching indexes:
        # get_exp_bins = lambda df, ref_chroms, _: (
        #     df.index.get_level_values("chrom").isin(ref_chroms).sum()
        # )
    elif contact_type == "trans":
        # that's what we expect as column names:
        expected_columns = ["region1", "region2", "n_valid", expected_name]
        # what would become a MultiIndex:
        expected_index = ["region1", "region2"]
        # expected dtype as a rudimentary form of validation:
        expected_dtype = {
            "region1": np.str,
            "region2": np.str,
            "n_valid": np.int64,
            expected_name: np.float64,
        }
        # # unique list of chroms mentioned in expected_path:
        # get_exp_chroms = lambda df: np.union1d(
        #     df.index.get_level_values("region1").unique(),
        #     df.index.get_level_values("region2").unique(),
        # )
        # # no way to get bins from trans-expected, so just get the number:
        # get_exp_bins = lambda _1, _2, correct_bins: correct_bins
    else:
        raise ValueError(
            "Incorrect contact_type: {}, ".format(contact_type),
            "Should have been caught by click.",
        )

    if min_dist < 0:
        min_diag = 3
    else:
        min_diag = int(np.ceil(min_dist / clr.binsize))

    if max_dist >= 0:
        max_diag = int(np.floor(max_dist / clr.binsize))
    else:
        max_diag = -1

    # use 'usecols' as a rudimentary form of validation,
    # and dtype. Keep 'comment' and 'verbose' - explicit,
    # as we may use them later:
    expected = pd.read_table(
        expected_path,
        usecols=expected_columns,
        index_col=expected_index,
        dtype=expected_dtype,
        comment=None,
        verbose=verbose,
    )

    # read bedGraph-file :
    track_columns = ["chrom", "start", "end", track_name]
    # specify dtype as a rudimentary form of validation:
    track_dtype = {
        "chrom": np.str,
        "start": np.int64,
        "end": np.int64,
        track_name: np.float64,
    }
    track = pd.read_table(
        track_path,
        usecols=track_columns,
        dtype=track_dtype,
        comment=None,
        verbose=verbose,
    )

    #############################################
    # CROSS-VALIDATE COOLER, EXPECTED AND TRACK:
    #############################################
    # TRACK vs COOLER:
    track_chroms = track["chrom"].unique()
    # We might want to try this eventually:
    # https://github.com/TMiguelT/PandasSchema
    # do simple column-name validation for now:
    if not set(track_chroms).issubset(clr.chromnames):
        raise ValueError(
            "Chromosomes in {} must be subset of ".format(track_path)
            + "chromosomes in cooler {}".format(cool_path)
        )
    # check number of bins:
    track_bins = len(track)
    cool_bins = clr.bins()[:]["chrom"].isin(track_chroms).sum()
    if not (track_bins == cool_bins):
        raise ValueError(
            "Number of bins is not matching: ",
            "{} in {}, and {} in {} for chromosomes {}".format(
                track_bins, track_path, cool_bins, cool_path, track_chroms
            ),
        )
    # # EXPECTED vs TRACK:
    # # validate expected a bit as well:
    # expected_chroms = get_exp_chroms(expected)
    # # do simple column-name validation for now:
    # if not set(track_chroms).issubset(expected_chroms):
    #     raise ValueError(
    #         "Chromosomes in {} must be subset of ".format(track_path)
    #         + "chromosomes in expected {}".format(expected_path)
    #     )
    # # and again bins are supposed to match up:
    # # only for cis though ...
    # expected_bins = get_exp_bins(expected, track_chroms, track_bins)
    # if not (track_bins == expected_bins):
    #     raise ValueError(
    #         "Number of bins is not matching: ",
    #         "{} in {}, and {} in {} for chromosomes {}".format(
    #             track_bins, track_path, expected_bins, expected_path, track_chroms
    #         ),
    #     )
    #############################################
    # CROSS-VALIDATION IS COMPLETE.
    #############################################

    track = saddle.mask_bad_bins((track, track_name), (clr.bins()[:], weight_name))

    if contact_type == "cis":
        getmatrix = saddle.make_cis_obsexp_fetcher(
            clr, (expected, expected_name), weight_name=weight_name
        )
    elif contact_type == "trans":
        getmatrix = saddle.make_trans_obsexp_fetcher(
            clr, (expected, expected_name), weight_name=weight_name
        )

    if quantiles:
        if len(range_):
            qlo, qhi = saddle.ecdf(track[track_name], range_)
        elif len(qrange):
            qlo, qhi = qrange
        else:
            qlo, qhi = 0.0, 1.0
        q_edges = np.linspace(qlo, qhi, n_bins)
        binedges = saddle.quantile(track[track_name], q_edges)
    else:
        if len(range_):
            lo, hi = range_
        elif len(qrange):
            lo, hi = saddle.quantile(track[track_name], qrange)
        else:
            lo, hi = track[track_name].min(), track[track_name].max()
        binedges = np.linspace(lo, hi, n_bins)

    digitized, hist = saddle.digitize_track(
        binedges, track=(track, track_name), regions=track_chroms
    )

    S, C = saddle.make_saddle(
        getmatrix,
        binedges,
        (digitized, track_name + ".d"),
        contact_type=contact_type,
        regions=regions,
        min_diag=min_diag,
        max_diag=max_diag,
    )

    saddledata = S / C

    to_save = dict(saddledata=saddledata, binedges=binedges, hist=hist)

    if strength:
        ratios = saddle.saddle_strength(S, C)
        ratios = ratios[1:-1]  # drop outlier bins
        to_save["saddle_strength"] = ratios

    # Save data
    np.savez(out_prefix + ".saddledump", **to_save)  # .npz auto-added
    digitized.to_csv(out_prefix + ".digitized.tsv", sep="\t", index=False)

    # Generate figure
    if len(fig):
        try:
            import matplotlib as mpl

            mpl.use("Agg")  # savefig only for now:
            import matplotlib.pyplot as plt
        except ImportError:
            print("Install matplotlib to use ", file=sys.stderr)
            sys.exit(1)

        if hist_color is None:
            color = (
                0.41568627450980394,
                0.8,
                0.39215686274509803,
            )  # sns.color_palette('muted')[2]
        else:
            color = mpl.colors.colorConverter.to_rgb(hist_color)
        title = op.basename(cool_path) + " ({})".format(contact_type)
        if quantiles:
            edges = q_edges
            track_label = track_name + " quantiles"
        else:
            edges = binedges
            track_label = track_name
        clabel = "(contact frequency / expected)"

        saddle.saddleplot(
            edges,
            hist,
            saddledata,
            scale=scale,
            vmin=vmin,
            vmax=vmax,
            color=color,
            title=title,
            xlabel=track_label,
            ylabel=track_label,
            clabel=clabel,
            cmap=cmap
        )

        for ext in fig:
            plt.savefig(out_prefix + "." + ext, bbox_inches="tight")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

from unittest.mock import Mock
MOCK_MODULES = [
    'cooltools.io.fastsavetxt',
    'cooltools.lib._numutils',
    'cooler',
    'cooler.core',
    'cooler.tools',
    'cython',
    'cytoolz',
    'dask',
    'h5py',
    'matplotlib',
    'matplotlib.cm',
    'matplotlib.pyplot',
    'numba',
    # 'numpy',
    'pandas',
    'scipy',
    'scipy.interpolate',
    'scipy.linalg',
    'scipy.sparse',
    'scipy.sparse.linalg',
    'scipy.ndimage',
    'scipy.ndimage.filters',
    'scipy.ndimage.interpolation',
    'scipy.signal',
    'scipy.stats',
    'sklearn',
    'sklearn.cluster',
]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()


# -- Project information -----------------------------------------------------

project = 'cooltools'
copyright = '2020, cooltoolers'
author = 'cooltoolers'


# -- General configuration ---------------------------------------------------

# Apparently readthedocs looks for contents.rst by default if this isn't set.
master_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx_click.ext',
    'recommonmark',
    'nbsphinx',
]

# Extension configuration
napoleon_google_docstring = False
# napoleon_use_param = False
# napoleon_use_ivar = True
napoleon_use_rtype = False

# Notebook prolog and epilog
nbsphinx_prolog = """"""
nbsphinx_epilog = r"""
----
{% set docname = env.doc2path(env.docname, base='docs') %}

This page was generated with nbsphinx_ from `{{ docname }}`__

__ https://github.com/open2c/cooltools/blob/master{{ env.config.release }}/{{ docname }}

.. _nbsphinx: https://nbsphinx.readthedocs.io/

"""

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Style overrides ----------------------------------------------------------
# Place CSS in _static directory
# def setup(app):
#     app.add_stylesheet('theme_overrides.css')



import click
import cooler

from . import cli
from .. import insulation


@cli.command()
@click.argument(
    "in_path",
    metavar="IN_PATH",
    type=str,
    nargs=1
)
@click.argument(
    "window",
    nargs=-1,
    metavar="WINDOW",
    type=int
)
@click.option(
    "--output",
    "-o",
    help="Specify output file name to store the insulation in a tsv format.",
    type=str,
    required=False,
)
@click.option(
    "--ignore-diags",
    help="The number of diagonals to ignore. By default, equals"
    " the number of diagonals ignored during IC balancing.",
    type=int,
    default=None,
    show_default=True,
)
@click.option(
    "--min-frac-valid-pixels",
    help="The minimal fraction of valid pixels in a sliding diamond. "
    "Used to mask bins during boundary detection.",
    type=float,
    default=0.66,
    show_default=True,
)
@click.option(
    "--min-dist-bad-bin",
    help="The minimal allowed distance to a bad bin. "
    "Used to mask bins during boundary detection.",
    type=int,
    default=0,
    show_default=True,
)
@click.option(
    "--window-pixels",
    help="If set then the window sizes are provided in units of pixels.",
    is_flag=True,
)
@click.option(
    "--append-raw-scores",
    help="Append columns with raw scores (sum_counts, sum_balanced, n_pixels) "
    "to the output table.",
    is_flag=True,
)
@click.option(
    "--chunksize",
    help="",
    type=int,
    default=20000000,
    show_default=True
)
@click.option(
    "--verbose",
    help="Report real-time progress.",
    is_flag=True
)
def diamond_insulation(
    in_path,
    window,
    output,
    ignore_diags,
    min_frac_valid_pixels,
    min_dist_bad_bin,
    window_pixels,
    append_raw_scores,
    chunksize,
    verbose,
):
    """
    Calculate the diamond insulation scores and call insulating boundaries.

    IN_PATH : The paths to a .cool file with a balanced Hi-C map.

    WINDOW : The window size for the insulation score calculations.
             Multiple space-separated values can be provided.
             By default, the window size must be provided in units of bp.
             When the flag --window-pixels is set, the window sizes must
             be provided in units of pixels instead.
    """

    clr = cooler.Cooler(in_path)
    if window_pixels:
        window = [win * clr.info["bin-size"] for win in window]

    ins_table = insulation.calculate_insulation_score(
        clr,
        window_bp=window,
        ignore_diags=ignore_diags,
        append_raw_scores=append_raw_scores,
        chunksize=chunksize,
        verbose=verbose,
    )

    ins_table = insulation.find_boundaries(
        ins_table,
        min_frac_valid_pixels=min_frac_valid_pixels,
        min_dist_bad_bin=min_dist_bad_bin,
    )

    # output to file if specified:
    if output:
        ins_table.to_csv(output, sep="\t", index=False, na_rep="nan")
    # or print into stdout otherwise:
    else:
        print(ins_table.to_csv(sep="\t", index=False, na_rep="nan"))



import warnings
import numpy as np
import pandas as pd
from .lib import peaks, numutils


def dirscore(pixels, bins, window=10, ignore_diags=2, balanced=True, signed_chi2=False):
    lo_bin_id = bins.index.min()
    hi_bin_id = bins.index.max() + 1
    N = hi_bin_id - lo_bin_id

    bad_bin_mask = (
        bins["weight"].isnull().values if balanced else np.zeros(N, dtype=bool)
    )

    diag_pixels = pixels[pixels["bin2_id"] - pixels["bin1_id"] <= (window - 1) * 2]
    if balanced:
        diag_pixels = diag_pixels[~diag_pixels["balanced"].isnull()]

    i = diag_pixels["bin1_id"].values - lo_bin_id
    j = diag_pixels["bin2_id"].values - lo_bin_id
    val = diag_pixels["balanced"].values if balanced else diag_pixels["count"].values

    sum_pixels_left = np.zeros(N)
    n_pixels_left = np.zeros(N)
    for i_shift in range(0, window):
        if i_shift < ignore_diags:
            continue

        mask = (i + i_shift == j) & (i + i_shift < N) & (j >= 0)
        sum_pixels_left += np.bincount(i[mask] + i_shift, val[mask], minlength=N)

        loc_bad_bin_mask = np.zeros(N, dtype=bool)
        if i_shift == 0:
            loc_bad_bin_mask |= bad_bin_mask
        else:
            loc_bad_bin_mask[i_shift:] |= bad_bin_mask[:-i_shift]
            loc_bad_bin_mask |= bad_bin_mask
        n_pixels_left[i_shift:] += 1 - loc_bad_bin_mask[i_shift:]

    sum_pixels_right = np.zeros(N)
    n_pixels_right = np.zeros(N)
    for j_shift in range(0, window):
        if j_shift < ignore_diags:
            continue

        mask = (i == j - j_shift) & (i < N) & (j - j_shift >= 0)

        sum_pixels_right += np.bincount(i[mask], val[mask], minlength=N)

        loc_bad_bin_mask = np.zeros(N, dtype=bool)
        loc_bad_bin_mask |= bad_bin_mask
        if j_shift == 0:
            loc_bad_bin_mask |= bad_bin_mask
        else:
            loc_bad_bin_mask[:-j_shift] |= bad_bin_mask[j_shift:]

        n_pixels_right[: (-j_shift if j_shift else None)] += (
            1 - loc_bad_bin_mask[: (-j_shift if j_shift else None)]
        )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        a = sum_pixels_left
        b = sum_pixels_right
        if signed_chi2:
            e = (a + b) / 2.0
            score = np.sign(b - a) * ((a - e) ** 2 + (b - e) ** 2) / e
        else:
            score = (b - a) / (a + b)

    return score


def _dirscore_dense(A, window=10, signed_chi2=False):
    N = A.shape[0]
    di = np.zeros(N)
    for i in range(0, N):
        lo = max(0, i - window)
        hi = min((i + window) + 1, N)
        b, a = np.nansum(A[i, i:hi]), np.nansum(A[i, lo : i + 1])
        if signed_chi2:
            e = (a + b) / 2.0
            if e:
                di[i] = np.sign(b - a) * ((a - e) ** 2 + (b - e) ** 2) / e
        else:
            di[i] = (b - a) / (a + b)
    mask = np.nansum(A, axis=0) == 0
    di[mask] = np.nan
    return di


def directionality(
    clr,
    window_bp=100000,
    balance="weight",
    min_dist_bad_bin=2,
    ignore_diags=None,
    chromosomes=None,
):
    """Calculate the diamond insulation scores and call insulating boundaries.

    Parameters
    ----------
    clr : cooler.Cooler
        A cooler with balanced Hi-C data.
    window_bp : int
        The size of the sliding diamond window used to calculate the insulation
        score.
    min_dist_bad_bin : int
        The minimal allowed distance to a bad bin. Do not calculate insulation
        scores for bins having a bad bin closer than this distance.
    ignore_diags : int
        The number of diagonals to ignore. If None, equals the number of
        diagonals ignored during IC balancing.

    Returns
    -------
    ins_table : pandas.DataFrame
        A table containing the insulation scores of the genomic bins and
        the insulating boundary strengths.
    """
    if chromosomes is None:
        chromosomes = clr.chromnames

    bin_size = clr.info["bin-size"]
    ignore_diags = (
        ignore_diags
        if ignore_diags is not None
        else clr._load_attrs(clr.root.rstrip("/") + "/bins/weight")["ignore_diags"]
    )
    window_bins = window_bp // bin_size

    if window_bp % bin_size != 0:
        raise Exception(
            "The window size ({}) has to be a multiple of the bin size {}".format(
                window_bp, bin_size
            )
        )

    dir_chrom_tables = []
    for chrom in chromosomes:
        chrom_bins = clr.bins().fetch(chrom)
        chrom_pixels = clr.matrix(as_pixels=True, balance=balance).fetch(chrom)

        # mask neighbors of bad bins
        is_bad_bin = np.isnan(chrom_bins["weight"].values)
        bad_bin_neighbor = np.zeros_like(is_bad_bin)
        for i in range(0, min_dist_bad_bin):
            if i == 0:
                bad_bin_neighbor = bad_bin_neighbor | is_bad_bin
            else:
                bad_bin_neighbor = bad_bin_neighbor | np.r_[[True] * i, is_bad_bin[:-i]]
                bad_bin_neighbor = bad_bin_neighbor | np.r_[is_bad_bin[i:], [True] * i]

        dir_chrom = chrom_bins[["chrom", "start", "end"]].copy()
        dir_chrom["bad_bin_masked"] = bad_bin_neighbor

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            dir_track = dirscore(
                chrom_pixels, chrom_bins, window=window_bins, ignore_diags=ignore_diags
            )
            dir_track[bad_bin_neighbor] = np.nan
            dir_track[~np.isfinite(dir_track)] = np.nan
            dir_chrom["directionality_ratio_{}".format(window_bp)] = dir_track

            dir_track = dirscore(
                chrom_pixels,
                chrom_bins,
                window=window_bins,
                ignore_diags=ignore_diags,
                signed_chi2=True,
            )
            dir_track[bad_bin_neighbor] = np.nan
            dir_track[~np.isfinite(dir_track)] = np.nan
            dir_chrom["directionality_index_{}".format(window_bp)] = dir_track

        dir_chrom_tables.append(dir_chrom)

    dir_table = pd.concat(dir_chrom_tables)
    return dir_table



# create a test for get_locally_adjusted_expected using mock dense matrices ...
# testing la_exp functions on a single piece of matrix
# no chunking applied here ...

import numpy as np
import pandas as pd

import os.path as op

# try importing stuff from dotfinder:
from cooltools.dotfinder import get_adjusted_expected_tile_some_nans

# adjust the path for data:
testdir = op.realpath(op.dirname(__file__))

# mock input data location:
mock_input = op.join(testdir, "data", "mock_inputs.npz")
mock_result = op.join(testdir, "data", "mock_res.csv.gz")

# load mock results:
mock_res = pd.read_csv(mock_result)
mock_res = mock_res.rename(columns={"row": "bin1_id", "col": "bin2_id"})

# load bunch of array from a numpy npz container:
arrays_loaded = np.load(mock_input)
# snippets of M_raw, M_ice, E_ice and v_ice are supposed
# to be there ...
mock_M_raw = arrays_loaded["mock_M_raw"]
mock_M_ice = arrays_loaded["mock_M_ice"]
mock_E_ice = arrays_loaded["mock_E_ice"]
mock_v_ice = arrays_loaded["mock_v_ice"]


# we need w-edge for tiling procedures:
# w = 3
# p = 1
# kernel type: 'donut'
# # just a simple donut kernel for testing:
kernel = np.array(
    [
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 0, 0, 0, 1, 1],
        [0, 0, 0, 0, 0, 0, 0],
        [1, 1, 0, 0, 0, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
        [1, 1, 1, 0, 1, 1, 1],
    ]
)
# bin size:
# mock data were extracted
# from 20kb matrix:
b = 20000
# 2MB aroud diagonal to look at:
band = 2e6


#################################################################
# main test function for new 'get_adjusted_expected_tile_some_nans':
#################################################################
def test_adjusted_expected_tile_some_nans():
    print("Running tile some nans la_exp test")
    # first, generate that locally-adjusted expected:
    # Ed_raw, mask_ndx, Cobs, Cexp, NN =
    res = get_adjusted_expected_tile_some_nans(
        origin=(0, 0),
        observed=mock_M_raw,  # should be RAW ...
        expected=mock_E_ice,
        bal_weights=mock_v_ice,
        kernels={"donut": kernel, "footprint": np.ones_like(kernel)},
    )
    # mock results are supposedly for
    # the contacts closer than the band
    # nucleotides to each other...
    # we want to retire that functional
    # from the 'get_adjusted_expected_tile_some_nans'
    # so we should complement that selection
    # here:
    nnans = 1  # no nans tolerated
    band_idx = int(band / b)
    is_inside_band = res["bin1_id"] > (res["bin2_id"] - band_idx)
    does_comply_nans = res["la_exp." + "footprint" + ".nnans"] < nnans
    # so, selecting inside band and nNaNs compliant results:
    res = res[is_inside_band & does_comply_nans].reset_index(drop=True)
    #
    # ACTUAL TESTS:
    # integer part of DataFrame must equals exactly:
    assert res[["bin1_id", "bin2_id"]].equals(mock_res[["bin1_id", "bin2_id"]])
    # compare floating point part separately:
    assert np.isclose(
        res["la_exp." + "donut" + ".value"], mock_res["la_expected"], equal_nan=True
    ).all()
    #
    # try recovering more NaNs ...
    # (allowing 1 here per pixel's footprint)...
    res = get_adjusted_expected_tile_some_nans(
        origin=(0, 0),
        observed=mock_M_raw,  # should be RAW...
        expected=mock_E_ice,
        bal_weights=mock_v_ice,
        kernels={"donut": kernel, "footprint": np.ones_like(kernel)},
    )
    # post-factum filtering resides
    # outside of get_la_exp now:
    nnans = 2  # just 1 nan tolerated
    band_idx = int(band / b)
    is_inside_band = res["bin1_id"] > (res["bin2_id"] - band_idx)
    does_comply_nans = res["la_exp." + "footprint" + ".nnans"] < nnans
    # so, selecting inside band and comply nNaNs results only:
    res = res[is_inside_band & does_comply_nans].reset_index(drop=True)

    # now we can only guess the size:
    assert len(res) > len(mock_res)



import os.path as op
import csv
import io
import click


class TabularFilePath(click.Path):
    def __init__(
        self, default_column_index, exists=False, resolve_path=False, allow_dash=False
    ):
        """
        Parameters
        ----------
        default_column : str or int
            Name of desired column or 0-based column index.
        exists : bool
        resolve_path : bool

        Returns
        -------
        path to file, column name or index

        """
        self.default_column_index = default_column_index
        super().__init__(
            exists=exists, resolve_path=resolve_path, allow_dash=allow_dash
        )

    def convert(self, value, param, ctx):
        if value is None:
            return
        file_path, _, field = value.partition("::")
        file_path = super().convert(file_path, param, ctx)
        if not field:
            col = self.default_column_index
        elif field.isdigit():
            col = int(field) - 1  # assume one-based from command line
            if col < 0:
                self.fail('Expected one-based column number, received "0".', param, ctx)
        else:
            col = field
        return file_path, col


def sniff_for_header(file_path, sep="\t", comment="#"):
    """
    Warning: reads the entire file into a StringIO buffer!

    """
    with open(file_path, "r") as f:
        buf = io.StringIO(f.read())

    sample_lines = []
    for line in buf:
        if not line.startswith(comment):
            sample_lines.append(line)
            break
    for _ in range(10):
        sample_lines.append(buf.readline())
    buf.seek(0)

    has_header = csv.Sniffer().has_header("\n".join(sample_lines))
    if has_header:
        names = sample_lines[0].strip().split(sep)
    else:
        names = None

    return buf, names


def validate_csv(ctx, param, value, default_column):
    if value is None:
        return
    file_path, _, field_name = value.partition("::")
    if not op.exists(file_path):
        raise click.BadParameter(
            'Path not found: "{}"'.format(file_path), ctx=ctx, param=param
        )
    if not field_name:
        field_name = default_column
    elif field_name.isdigit():
        field_name = int(field_name)
    return file_path, field_name



import multiprocess as mp
import pandas as pd
from itertools import combinations
import cooler
from bioframe import parse_regions
from .. import expected

import click
from . import cli
from . import util

# might be relevant to us ...
# https://stackoverflow.com/questions/46577535/how-can-i-run-a-dask-distributed-local-cluster-from-the-command-line
# http://distributed.readthedocs.io/en/latest/setup.html#using-the-command-line


@cli.command()
@click.argument("cool_path", metavar="COOL_PATH", type=str, nargs=1)
@click.option(
    "--nproc", "-p",
    help="Number of processes to split the work between."
    "[default: 1, i.e. no process pool]",
    default=1,
    type=int,
)
@click.option(
    "--chunksize", "-c",
    help="Control the number of pixels handled by each worker process at a time.",
    type=int,
    default=int(10e6),
    show_default=True,
)
@click.option(
    "--output", "-o",
    help="Specify output file name to store the expected in a tsv format.",
    type=str,
    required=False,
)
@click.option(
    "--hdf",
    help="Use hdf5 format instead of tsv."
    " Output file name must be specified [Not Implemented].",
    is_flag=True,
    default=False,
)
@click.option(
    "--contact-type", "-t",
    help="compute expected for cis or trans region of a Hi-C map."
    "trans-expected is calculated for pairwise combinations of specified regions.",
    type=click.Choice(["cis", "trans"]),
    default="cis",
    show_default=True,
)
@click.option(
    "--regions",
    help="Path to a BED file containing genomic regions "
    "for which expected will be calculated. Region names can"
    "be provided in a 4th column, otherwise UCSC notaion is used."
    "When not specified, expected is calculated for all chromosomes",
    type=click.Path(exists=True),
    required=False,
)
@click.option(
    "--balance/--no-balance",
    help="Apply balancing weights to data before calculating expected."
    "Bins masked in the balancing weights are ignored from calcualtions.",
    is_flag=True,
    default=True,
    show_default=True,
)
@click.option(
    "--weight-name",
    help="Use balancing weight with this name.",
    type=str,
    default="weight",
    show_default=True,
)
@click.option(
    "--blacklist",
    help="Path to a 3-column BED file containing genomic regions to mask "
    "out during calculation of expected. Overwrites inference of "
    "'bad' regions from balancing weights. [Not Implemented]",
    type=click.Path(exists=True),
    required=False,
)
@click.option(
    "--ignore-diags",
    help="Number of diagonals to neglect for cis contact type",
    type=int,
    default=2,
    show_default=True,
)
def compute_expected(
    cool_path,
    nproc,
    chunksize,
    output,
    hdf,
    contact_type,
    regions,
    balance,
    weight_name,
    blacklist,
    ignore_diags,
):
    """
    Calculate expected Hi-C signal either for cis or for trans regions
    of chromosomal interaction map.

    When balancing weights are not applied to the data, there is no
    masking of bad bins performed.

    COOL_PATH : The paths to a .cool file with a balanced Hi-C map.

    """
    if blacklist is not None:
        raise NotImplementedError(
            "Custom genomic regions for masking from calculation of expected"
            "are not implemented."
        )
        # use blacklist-ing from cooler balance module
        # https://github.com/mirnylab/cooler/blob/843dadca5ef58e3b794dbaf23430082c9a634532/cooler/cli/balance.py#L175

    clr = cooler.Cooler(cool_path)
    if regions is None:
        regions = [(chrom, 0, clr.chromsizes[chrom]) for chrom in clr.chromnames]
        regions = parse_regions(regions)
        regions["name"] = clr.chromnames
    else:
        regions_buf, names = util.sniff_for_header(regions)
        regions = pd.read_csv(regions_buf, sep="\t", header=None)
        if regions.shape[1] not in (3, 4):
            raise ValueError(
                "The region file does not have three or four tab-delimited columns."
                "We expect a bed file with columns chrom, start, end, and optional name"
            )
        if regions.shape[1] == 4:
            regions = regions.rename(columns={0:"chrom",1:"start",2:"end",3:"name"})
            regions = parse_regions(regions)
        else:
            regions = regions.rename(columns={0:"chrom",1:"start",2:"end"})
            regions["name"] = list(regions.apply(lambda x: "{}:{}-{}".format(*x), axis=1))
            regions = parse_regions(regions)

    # define transofrms - balanced and raw ('count') for now
    if balance:
        weight1 = weight_name + "1"
        weight2 = weight_name + "2"
        transforms = {"balanced": lambda p: p["count"] * p[weight1] * p[weight2]}
    else:
        # no masking bad bins of any kind, when balancing is not applied
        weight_name = None
        transforms = {}

    # execution details
    if nproc > 1:
        pool = mp.Pool(nproc)
        map_ = pool.map
    else:
        map_ = map

    # using try-clause to close mp.Pool properly
    try:
        if contact_type == "cis":
            result = expected.diagsum(
                clr,
                regions,
                transforms=transforms,
                weight_name=weight_name,
                bad_bins=None,
                chunksize=chunksize,
                ignore_diags=ignore_diags,
                map=map_,
            )
        elif contact_type == "trans":
            # prepare pairwise combinations of regions for trans-expected (blocksum):
            regions_pairwise = combinations(regions.itertuples(index=False), 2)
            regions1, regions2 = zip(*regions_pairwise)
            result = expected.blocksum_asymm(
                clr,
                regions1 = pd.DataFrame(regions1),
                regions2 = pd.DataFrame(regions2),
                transforms=transforms,
                weight_name=weight_name,
                bad_bins=None,
                chunksize=chunksize,
                map=map_,
            )
    finally:
        if nproc > 1:
            pool.close()

    # calculate actual averages by dividing sum by n_valid:
    result["count.avg"] = result["count.sum"] / result["n_valid"]
    for key in transforms.keys():
        result[key + ".avg"] = result[key + ".sum"] / result["n_valid"]

    # output to file if specified:
    if output:
        result.to_csv(output, sep="\t", index=False, na_rep="nan")
    # or print into stdout otherwise:
    else:
        print(result.to_csv(sep="\t", index=False, na_rep="nan"))

    # would be nice to have some binary output to preserve precision.
    # to_hdf/read_hdf should work in this case as the file is small .
    # still debated as to how should we store it - store in cooler seems
    # to be consensus:
    if hdf:
        raise NotImplementedError("hdf output is to be implemented")



from collections import defaultdict
import numpy as np
import pandas as pd

# from scipy.sparse import coo_matrix
from cooler.core import _IndexingMixin


def arg_prune_partition(seq, step):
    """
    Take a monotonic sequence of integers and downsample it such that they
    are at least ``step`` apart (roughly), preserving the first and last
    elements. Returns indices, not values.

    """
    lo, hi = seq[0], seq[-1]
    num = 2 + (hi - lo) // step
    cuts = np.linspace(lo, hi, num, dtype=int)
    return np.unique(np.searchsorted(seq, cuts))


class CSRSelector(_IndexingMixin):
    """
    Instantiates 2D range queries.

    Example
    -------
    >>> selector = CSRSelector(h5, (100, 100), 'count', 10000)
    >>> query = selector[lo1:hi1, lo2:hi2]

    """

    def __init__(self, grp, shape, field, chunksize):
        self.grp = grp
        self.shape = shape
        self.field = field
        self.chunksize = chunksize
        self.offset_selector = grp["indexes"]["bin1_offset"]
        self.bin1_selector = grp["pixels"]["bin1_id"]
        self.bin2_selector = grp["pixels"]["bin2_id"]
        self.data_selector = grp["pixels"][field]

    def _make_getchunk(self, ispan, jspan):
        # Factory for function that executes any piece of a 2D range query by
        # index.

        bin1_selector = self.bin1_selector
        bin2_selector = self.bin2_selector
        data_selector = self.data_selector
        field = self.field
        i0, i1 = ispan
        j0, j1 = jspan

        # coarsegrain the offsets to extract a big chunk of rows at a time
        if (i1 - i0 < 1) or (j1 - j0 < 1):
            offsets = []
            loc_pruned_offsets = []
        else:
            offsets = self.offset_selector[i0 : i1 + 1]
            loc_pruned_offsets = arg_prune_partition(offsets, self.chunksize)

        self._loc_pruned_offsets = loc_pruned_offsets
        # i0 -- matrix row number offset
        # o0 -- corresponding pixel id offset = offsets[0]

        # let's take the downsampled subset of pixel id offsets [o0, ...., o1]
        # each successive pair corresponds to a "piece" of the query
        def getchunk(chunk_id, include_index=False):
            out = {"bin1_id": [], "bin2_id": [], field: []}
            if include_index:
                out["__index"] = []

            # extract a chunk of on-disk rows
            oi, of = loc_pruned_offsets[chunk_id], loc_pruned_offsets[chunk_id + 1]
            p0, p1 = offsets[oi], offsets[of]
            slc = slice(p0, p1)

            bin2_extracted = bin2_selector[slc]
            data_extracted = data_selector[slc]
            if include_index:
                ind_extracted = np.arange(slc.start, slc.stop)

            # go row by row and filter
            for i in range(oi, of):
                # correct the offsets
                lo = offsets[i] - p0
                hi = offsets[i + 1] - p0

                # this row
                bin2 = bin2_extracted[lo:hi]

                # filter for the range of j values we want
                mask = (bin2 >= j0) & (bin2 < j1)
                cols = bin2[mask]

                # apply same mask for data
                data = data_extracted[lo:hi][mask]

                # shortcut for row data
                rows = np.full(len(cols), i0 + i, dtype=bin1_selector.dtype)

                out["bin1_id"].append(rows)
                out["bin2_id"].append(cols)
                out[field].append(data)
                if include_index:
                    out["__index"].append(ind_extracted[lo:hi][mask])

            if len(out):
                for k in out.keys():
                    out[k] = np.concatenate(out[k], axis=0)
            else:
                out["bin1_id"] = np.array([], dtype=bin1_selector.dtype)
                out["bin2_id"] = np.array([], dtype=bin2_selector.dtype)
                out[field] = np.array([], dtype=data_selector.dtype)
                if include_index:
                    out["__index"] = np.array([], dtype=np.int64)

            return out

        return getchunk, loc_pruned_offsets

    def __getitem__(self, key):
        s1, s2 = self._unpack_index(key)
        ispan = self._process_slice(s1, self.shape[0])
        jspan = self._process_slice(s2, self.shape[1])
        getchunk, loc_pruned_offsets = self._make_getchunk(ispan, jspan)
        return RangeQuery(self, ispan, jspan, self.field, getchunk, loc_pruned_offsets)


class RangeQuery(object):
    """
    Executor that fulfills a partitioned 2D range query using a variety of outputs.

    """

    def __init__(self, selector, ispan, jspan, field, getchunk, loc_pruned_offsets):
        self.selector = selector
        self.ispan = ispan
        self.jspan = jspan
        self.field = field
        self.n_chunks = len(loc_pruned_offsets) - 1
        self._locs = loc_pruned_offsets
        self._getchunk = getchunk

    def read_chunk(self, i, include_index=False):
        """Read any chunk of the partitioned query as a dictionary."""
        if not 0 <= i < self.n_chunks:
            raise IndexError(i)
        return self._getchunk(i, include_index)

    def read_chunked(self, include_index=False):
        """Iterator over chunks (as dictionaries)."""
        for i in range(self.n_chunks):
            yield self._getchunk(i, include_index)

    def read(self, include_index=False):
        """Read the complete range query as a dictionary"""
        result = list(self.read_chunked(include_index))
        return {
            k: np.concatenate([d[k] for d in result], axis=0)
            for k in ["bin1_id", "bin2_id", self.field]
        }

    def __repr__(self):
        return (
            "{self.__class__.__name__}"
            '({self.ispan}, {self.jspan}, "{self.field}", ...) '
            "[{n} piece(s)]"
        ).format(self=self, n=self.n_chunks)



# -*- coding: utf-8 -*-
import numpy as np


def isrle(starts, lengths, values):
    if not (len(starts) == len(lengths) == len(values)):
        return False

    if np.any(np.diff(starts) < 0):
        return False

    ends = starts + lengths
    if np.any(ends[:-1] > starts[1:]):
        return False

    return True


def rlencode(x, dropna=False):
    """
    Run length encoding.
    Based on http://stackoverflow.com/a/32681075, which is based on the rle
    function from R.

    Parameters
    ----------
    x : 1D array_like
        Input array to encode
    dropna: bool, optional
        Drop all runs of NaNs.

    Returns
    -------
    start positions, run lengths, run values

    """
    where = np.flatnonzero
    x = np.asarray(x)
    n = len(x)
    if n == 0:
        return (np.array([], dtype=int),
                np.array([], dtype=int),
                np.array([], dtype=x.dtype))

    isnumeric = np.issubdtype(x.dtype, np.number)

    if isnumeric:
        starts = np.r_[0, where(~np.isclose(x[1:], x[:-1], equal_nan=True)) + 1]
    else:
        starts = np.r_[0, where(x[1:] != x[:-1]) + 1]
    lengths = np.diff(np.r_[starts, n])
    values = x[starts]

    if isnumeric and dropna:
        mask = ~np.isnan(values)
        starts, lengths, values = starts[mask], lengths[mask], values[mask]

    return starts, lengths, values


def rldecode(starts, lengths, values, minlength=None):
    """
    Decode a run-length encoding of a 1D array.

    Parameters
    ----------
    starts, lengths, values : 1D array_like
        The run-length encoding.
    minlength : int, optional
        Minimum length of the output array.

    Returns
    -------
    1D array. Missing data will be filled with NaNs.

    """
    starts, lengths, values = map(np.asarray, (starts, lengths, values))
    # TODO: check validity of rle
    ends = starts + lengths
    n = ends[-1]
    if minlength is not None:
        n = max(minlength, n)
    x = np.full(n, np.nan)
    for lo, hi, val in zip(starts, ends, values):
        x[lo:hi] = val
    return x


def iterruns(x, value=None, **kwargs):
    starts, lengths, values = rlencode(x, **kwargs)
    if value is None:
        ends = starts + lengths
        return zip(starts, ends, values)
    else:
        mask = values == value
        starts, lengths = starts[mask], lengths[mask]
        ends = starts + lengths
        return zip(starts, ends)


def fillgaps(starts, lengths, values, minlength=None, fill_value=np.nan):
    """
    Add additional runs to fill in spaces between runs. Defaults to runs of NaN.
    """
    where = np.flatnonzero
    n = starts[-1] + lengths[-1]
    if minlength is not None:
        n = max(minlength, n)

    ends = starts + lengths
    lo = np.r_[0, ends]
    hi = np.r_[starts, n]
    gap_locs = where((hi - lo) > 0)
    if len(gap_locs):
        starts = np.insert(starts, gap_locs, lo[gap_locs])
        lengths = np.insert(lengths, gap_locs, hi[gap_locs] - lo[gap_locs])
        values = np.insert(values, gap_locs, fill_value)
    return starts, lengths, values


def dropgaps(starts, lengths, values):
    """
    Discard runs of NaN.
    """
    mask = np.isnan(values)
    starts, lengths, values = starts[mask], lengths[mask], values[mask]
    return starts, lengths, values


def align(slv1, slv2, minlength=None):
    """
    Remove NaN runs and runs of length zero and stich together consecutive runs
    of the same value.

    """
    starts1, lengths1, values1 = fillgaps(*slv1)
    starts2, lengths2, values2 = fillgaps(*slv2)
    n1 = starts1[-1] + lengths1[-1]
    n2 = starts2[-1] + lengths2[-1]
    if minlength is not None:
        n = max(minlength, n1, n2)

    starts = np.concatenate([starts1, starts2])
    values = np.concatenate([values1, values2])
    idx = np.argsort(starts)
    starts = starts[idx]
    values = values[idx]
    lengths = np.diff(np.r_[starts, n])
    return starts, lengths, values


def simplify(starts, lengths, values, minlength=None):
    """
    Remove NaN runs and runs of length zero and stich together consecutive runs
    of the same value.

    """
    starts, lengths, values = fillgaps(starts, lengths, values, minlength)
    n = starts[-1] + lengths[-1]

    is_nontrivial = lengths > 0
    starts = starts[is_nontrivial]
    values = values[is_nontrivial]

    is_new_run = np.r_[True, ~np.isclose(values[:-1], values[1:], equal_nan=True)]
    starts = starts[is_new_run]
    values = values[is_new_run]

    lengths = np.r_[starts[1:] - starts[:-1], n - starts[-1]]

    mask = ~np.isnan(values)
    return starts[mask], lengths[mask], values[mask]



import sys
import click
from . import cli


@cli.group()
def genome():
    """
    Utilities for binned genome assemblies.

    """


@genome.command()
@click.argument("db")
def fetch_chromsizes(db):
    import bioframe

    chromsizes = bioframe.fetch_chromsizes(db)
    print(chromsizes.to_csv(sep="\t"))


@genome.command()
@click.argument("chromsizes_path")
@click.argument("binsize", type=int)
@click.option(
    "--all-names",
    help="Parse all chromosome names from file, not only default r\"^chr[0-9]+$\", r\"^chr[XY]$\", r\"^chrM$\". ",
    is_flag=True
)
def binnify(chromsizes_path, binsize, all_names):
    import bioframe

    chromsizes = bioframe.read_chromsizes(chromsizes_path, filter_chroms=not(all_names))
    bins = bioframe.binnify(chromsizes, binsize)
    print(bins.to_csv(sep="\t", index=False))


@genome.command()
@click.argument("chromsizes_path")
@click.argument("fasta_path")
@click.argument("enzyme_name")
def digest(chromsizes_path, fasta_path, enzyme_name):
    import bioframe

    chromsizes = bioframe.read_chromsizes(chromsizes_path, all_names=True)
    fasta_records = bioframe.load_fasta(fasta_path, engine="pyfaidx", as_raw=True)
    if not chromsizes.index.isin(fasta_records).all():
        raise ValueError(
            "Some chromosomes mentioned in {}"
            " are not found in {}".format(chromsizes_path, fasta_path)
        )
    frags = bioframe.digest(fasta_records, enzyme_name)
    print(frags.to_csv(sep="\t", index=False))


@genome.command()
@click.argument("bins_path")
@click.argument("fasta_path")
@click.option("--mapped-only", is_flag=True, default=True)
def gc(bins_path, fasta_path, mapped_only):
    import bioframe
    import pandas as pd

    if bins_path == "-":
        bins_path = sys.stdin
    bins = pd.read_table(bins_path)
    chromosomes = bins["chrom"].unique()
    fasta_records = bioframe.load_fasta(fasta_path, engine="pyfaidx", as_raw=True)
    if any(chrom not in fasta_records.keys() for chrom in chromosomes):
        raise ValueError(
            "Some chromosomes mentioned in {}"
            " are not found in {}".format(bins_path, fasta_path)
        )
    bins = bioframe.frac_gc(bins, fasta_records, mapped_only)
    print(bins.to_csv(sep="\t", index=False))


@genome.command()
@click.argument("bins_path")
@click.argument("db")
def genecov(bins_path, db):
    """
    BINS_PATH is the path to bintable.
    
    DB is the name of the genome assembly.
    The gene locations will be automatically downloaded from teh UCSC goldenPath.
    """
    import bioframe
    import pandas as pd

    bins = pd.read_table(bins_path)
    bins = bioframe.frac_gene_coverage(bins, db)
    print(bins.to_csv(sep="\t", index=False))



import numpy as np
import cooler.tools


def _zero_diags(chunk, n_diags):
    if n_diags > 0:
        mask = np.abs(chunk["pixels"]["bin1_id"] - chunk["pixels"]["bin2_id"]) < n_diags
        chunk["pixels"]["count"][mask] = 0

    return chunk


def _get_chunk_coverage(chunk, pixel_weight_key="count"):
    """
    Compute cis and total coverages of a cooler chunk.
    
    Parameters
    ----------
    chunk : dict of dict/pd.DataFrame
        A cooler chunk produced by the cooler split-apply-combine pipeline.
    pixel_weight_key: str
        The key of a pixel chunk to retrieve pixel weights.
    
    Returns
    -------
    covs : np.array 2 x n_bins    
        A numpy array with cis (the first row) and total (the 4nd) coverages.
    """

    bins = chunk["bins"]
    pixels = chunk["pixels"]
    n_bins = len(bins["chrom"])
    covs = np.zeros((2, n_bins))
    pixel_weights = pixels[pixel_weight_key]

    cis_mask = bins["chrom"][pixels["bin1_id"]] == bins["chrom"][pixels["bin2_id"]]
    covs[0] += np.bincount(
        pixels["bin1_id"], weights=pixel_weights * cis_mask, minlength=n_bins
    )
    covs[0] += np.bincount(
        pixels["bin2_id"], weights=pixel_weights * cis_mask, minlength=n_bins
    )

    covs[1] += np.bincount(pixels["bin1_id"], weights=pixel_weights, minlength=n_bins)
    covs[1] += np.bincount(pixels["bin2_id"], weights=pixel_weights, minlength=n_bins)

    covs = covs / 2

    return covs


def get_coverage(
    clr,
    ignore_diags=None,
    chunksize=int(1e7),
    map=map,
    use_lock=False,
    store=False,
    store_names=["cis_raw_cov", "tot_raw_cov"],
):

    """
    Calculate the sums of cis and genome-wide contacts (aka coverage aka marginals) for 
    a sparse Hi-C contact map in Cooler HDF5 format. 

    Parameters
    ----------
    clr : cooler.Cooler
        Cooler object
    chunksize : int, optional
        Split the contact matrix pixel records into equally sized chunks to
        save memory and/or parallelize. Default is to use all the pixels at
        once.
    map : callable, optional
        Map function to dispatch the matrix chunks to workers.
        Default is the builtin ``map``, but alternatives include parallel map
        implementations from a multiprocessing pool.
    ignore_diags : int, optional
        Drop elements occurring on the first ``ignore_diags`` diagonals of the
        matrix (including the main diagonal). 
        If None, equals the number of diagonals ignored during IC balancing.
    store : bool, optional
        If True, store the results in the file when finished. Default is False.
    store_names : list, optional
        Names of the columns of the bin table to save cis and total coverages. 
    
    Returns
    -------
    cis_cov : 1D array, whose shape is the number of bins in ``h5``. Vector of bin sums in cis.
    tot_cov : 1D array, whose shape is the number of bins in ``h5``. Vector of bin sums.

    """

    try:
        ignore_diags = (
            ignore_diags
            if ignore_diags is not None
            else clr._load_attrs(clr.root.rstrip("/") + "/bins/weight")["ignore_diags"]
        )
    except:
        raise ValueError(
            "Please, specify ignore_diags and/or IC balance this cooler! Cannot access the value used in IC balancing. "
        )

    chunks = cooler.tools.split(clr, chunksize=chunksize, map=map, use_lock=use_lock)

    if ignore_diags:
        chunks = chunks.pipe(_zero_diags, n_diags=ignore_diags)

    n_bins = clr.info["nbins"]
    covs = chunks.pipe(_get_chunk_coverage).reduce(np.add, np.zeros((2, n_bins)))

    if store:
        with clr.open("r+") as grp:
            for store_name, cov_arr in zip(store_names, covs):
                if store_name in grp["bins"]:
                    del grp["bins"][store_name]
                h5opts = dict(compression="gzip", compression_opts=6)
                grp["bins"].create_dataset(store_name, data=cov_arr, **h5opts)

    return covs



from functools import partial, reduce
from multiprocess import Pool
from operator import add

import numpy as np
import pandas
import pandas
import h5py

from scipy.sparse import linalg
from cooler.tools import split, partition
import cooler


def bnewt(matvec, mask, tol=1e-6, x0=None, delta=0.1, Delta=3, fl=0):
    """
    A balancing algorithm for symmetric matrices

    X = BNEWT(A) attempts to find a vector X such that
    diag(X)*A*diag(X) is close to doubly stochastic. A must
    be symmetric and nonnegative.

    Parameters
    ----------
    matvec : callable
        Linear operator that returns the matrix-vector product with x
    mask : 1D array of bool
        Mask of good bins
    tol : float
        Error tolerance
    x0 : 1D array
        Initial guess
    delta : float
        How close balancing vectors can get to the edge of the positive cone
    Delta : float
        How far balancing vectors can get from the edge of the positive cone

    We use a relative measure on the size of elements.

    Returns
    -------
    x : 1D array
        balancing weights
    res : float
        residual error, measured by norm(diag(x)*A*x - e)

    """
    # Initialize
    n = mask.sum()

    e = np.ones(n)
    if x0 is None:
        x0 = e.copy()
    res = []

    # Inner stopping criterion parameters.
    g = 0.9
    etamax = 0.1
    eta = etamax
    stop_tol = tol * 0.5
    x = x0
    rt = tol ** 2
    v = x * matvec(x, mask)

    rk = 1 - v
    rho_km1 = np.dot(rk, rk)
    rout = rho_km1
    rold = rout

    MVP = 0  # We’ll count matrix vector products.
    i = 0  # Outer iteration count.

    if fl == 1:
        print("it in. it res", flush=True)

    # Outer iteration
    while rout > rt:
        i += 1
        k = 0
        y = e.copy()
        innertol = max((eta ** 2) * rout, rt)

        # Inner iteration by Conjugate Gradient
        while rho_km1 > innertol:
            k += 1

            if k == 1:
                Z = rk / v
                p = Z.copy()
                rho_km1 = np.dot(rk, Z)
            else:
                beta = rho_km1 / rho_km2
                p = Z + beta * p

            # Update search direction efficiently.
            w = x * matvec(x * p, mask) + v * p

            alpha = rho_km1 / np.dot(p, w)
            ap = alpha * p

            # Test distance to boundary of cone.
            ynew = y + ap
            if min(ynew) <= delta:
                if delta == 0:
                    break
                idx = ap < 0
                gamma = np.min((delta - y[idx]) / ap[idx])
                y = y + gamma * ap
                break

            if max(ynew) >= Delta:
                idx = ynew > Delta
                gamma = np.min((Delta - y[idx]) / ap[idx])
                y = y + gamma * ap
                break

            y = ynew.copy()
            rk = rk - alpha * w
            rho_km2 = rho_km1
            Z = rk / v
            rho_km1 = np.dot(rk, Z)

        x = x * y
        v = x * matvec(x, mask)

        rk = 1 - v
        rho_km1 = np.dot(rk, rk)
        rout = rho_km1
        MVP += k + 1

        # Update inner iteration stopping criterion.
        rat = rout / rold
        rold = rout
        res_norm = np.sqrt(rout)
        eta_o = eta
        eta = g * rat
        if g * (eta_o ** 2) > 0.1:
            eta = max(eta, g * (eta_o ** 2))

        eta = max(min(eta, etamax), stop_tol / res_norm)
        if fl == 1:
            print("%3d\t%6d\t%.3e" % (i, k, res_norm), flush=True)
        res.append(res_norm)

        print("Matrix-vector products = %6d" % (MVP,), flush=True)

    x_full = np.zeros(len(mask))
    x_full[mask] = x
    return x_full, np.array(res)



from operator import add
from cooler.tools import split
import cython
import numpy as np

cimport numpy as np
from cpython cimport bool


ctypedef unsigned long ulong
ctypedef unsigned int uint
ctypedef unsigned short ushort
ctypedef unsigned char uchar

cdef extern from "stdlib.h":
    long c_libc_random "random"()
    double c_libc_drandom "drand48"()


def logbins(lo, hi, ratio=0, N=0, version=2):
    """Make bins with edges evenly spaced in log-space.

    Parameters
    ----------
    lo, hi : int
        The span of the bins.
    ratio : float
        The target ratio between the upper and the lower edge of each bin.
        Either ratio or N must be specified.
    N : int
        The target number of bins. The resulting number of bins is not guaranteed.
        Either ratio or N must be specified.

    version : int
        version=1 is a legacy mirnylib version

        version=2 fixes ordering of bin gaps for some values, making sure
        that bin gaps are always increasing (2/2/2020)

    """
    lo = int(lo)
    hi = int(hi)
    if ratio != 0:
        if N != 0:
            raise ValueError("Please specify N or ratio")
        N = int(np.log(hi / lo) / np.log(ratio))
    elif N == 0:
        raise ValueError("Please specify N or ratio")
    data10 = np.logspace(np.log10(lo), np.log10(hi), N)
    data10 = np.array(np.rint(data10), dtype=int)
    data10 = np.sort(np.unique(data10))

    if version > 1:
        data10 = np.cumsum(np.r_[data10[0], np.sort(np.diff(data10))])

    assert data10[0] == lo
    assert data10[-1] == hi

    return data10


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def observed_over_expected(matrix, mask=None):
    "Calculates observedOverExpected of any contact map, over regions where mask==1"

    cdef int N = matrix.shape[0]

    cdef np.ndarray[np.double_t, ndim = 2] data = np.array(
        matrix, dtype = np.double, order = "C")
    cdef np.ndarray[np.double_t, ndim = 2] datamask
    cdef bool has_mask = mask is not None
    if has_mask:
        datamask = np.array(mask==1, dtype = np.double, order = "C")

    cdef i, j, offset, lo, hi, n_pixels
    cdef double sum_pixels, mean_pixel
    n_pixels = 0
    sum_pixels = 0

    bins = [0] + list(logbins(1, N, 1.03))
    cdef int n_bins = len(bins)
    for lo, hi in zip(bins[:n_bins-1], bins[1:]):
        sum_pixels = 0
        n_pixels = 0
        for offset in range(lo, hi):
            for j in range(0, N-offset):
                if not has_mask or datamask[offset+j, j]==1:
                    sum_pixels += data[offset+j, j]
                    n_pixels += 1
        #print start, end, count
        mean_pixel = sum_pixels / n_pixels
        if mean_pixel != 0:
            for offset in range(lo, hi):
                for j in range(0, N-offset):
                    if not has_mask or datamask[offset+j, j]==1:
                        data[offset + j, j] /= mean_pixel
                        if offset > 0:
                            data[j, offset+j] /= mean_pixel

    return data


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterative_correction_symmetric(
    x, max_iter=1000, ignore_diags = 0, tolerance=1e-5):
    """The main method for correcting DS and SS read data.
    By default does iterative correction, but can perform an M-time correction

    Parameters
    ----------

    x : np.ndarray
        A symmetric matrix to correct.
    max_iter : int
        The maximal number of iterations to take.
    ignore_diags : int
        The number of diagonals to ignore during iterative correction.
    tolerance : float
        If less or equal to zero, will perform max_iter iterations.

    """
    cdef int N = len(x)

    x = np.array(x, np.double, order = 'C')
    cdef np.ndarray[np.double_t, ndim = 2] _x = x
    cdef np.ndarray[np.double_t, ndim = 1] s
    cdef np.ndarray[np.double_t, ndim = 1] s0
    cdef np.ndarray[np.double_t, ndim = 1] totalBias = np.ones(N, np.double)

    cdef int i, j, iternum
    cdef bool converged = False

    for iternum in range(max_iter):
        s0 = np.sum(_x, axis = 1)

        mask = (s0 == 0)

        s = s0.copy()
        for diag_idx in range(ignore_diags):   #excluding the diagonal
            if diag_idx == 0:
                s -= np.diagonal(_x)
            else:
                dia = np.array(np.diagonal(_x, diag_idx))
                s[diag_idx:] = s[diag_idx:] - dia
                s[:len(s) - diag_idx] = s[:len(s) - diag_idx] - dia

        s = s / np.mean(s[s0 != 0])
        s[s0==0] = 1
        s -= 1
        s *= 0.8
        s += 1
        totalBias *= s

        for i in range(N):
            for j in range(N):
                _x[i, j] = _x[i, j] / (s[i] * s[j])

        if (tolerance > 0) and (np.abs(s - 1).max() < tolerance):
            converged=True
            break

    corr = totalBias[s0!=0].mean()  #mean correction factor
    x  = x * corr * corr #renormalizing everything
    totalBias /= corr
    report = {'converged':converged, 'iternum':iternum}

    return x, totalBias, report


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.wraparound(False)
def fake_cis(
        np.ndarray[np.double_t, ndim=2] data,
        np.ndarray[np.uint8_t, ndim=2] mask
):
    """
    Replace target pixels (e.g. cis areas) of a contact matrix with decoy data
    sampled from source pixels (e.g. trans areas) in the same row or column.

    Operation is done in-place.

    Parameters
    ----------
    data : array of double (n, n)
        Contact matrix.
    mask : array of uint8 (n, n)
        Ternary mask. 0 = source pixels to sample from, 1 = target pixels to
        fill, 2 = invalid pixels to ignore.

    """
    cdef int N = data.shape[0]
    cdef int i, j, row_or_col, src_i, src_j

    for i in range(N):
        for j in range(i, N):

            if mask[i, j] == 1:  # valid target pixel?
                while True:
                    row_or_col = c_libc_random() % 2
                    if row_or_col == 0:
                        src_i = i
                    else:
                        src_i = j
                    src_j = c_libc_random() % N

                    if mask[src_i, src_j] == 0:  # valid source pixel?
                        data[i, j] = data[j, i] = data[src_i, src_j]
                        break


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
def _matvec_sparse_symmetric(
        np.ndarray[np.double_t, ndim=1] y,
        np.ndarray[np.int_t, ndim=1] bin1,
        np.ndarray[np.int_t, ndim=1] bin2,
        np.ndarray[np.double_t, ndim=1] values,
        np.ndarray[np.double_t, ndim=1] x):
    """
    Perform matrix vector product A * x for a sparse real
    symmetric matrix A.

    bin1, bin2 : 1D array
        sparse representation of A
    x : 1D array
        input vector
    y : 1D array
        output vector (values are added to this)

    """
    cdef int n = bin1.shape[0]
    cdef int ptr, i, j
    cdef double Aij

    for ptr in range(n):
        i, j, Aij = bin1[ptr], bin2[ptr], values[ptr]
        y[i] += (Aij * x[j])
        if j > i:
            y[j] += (Aij * x[i])


def _matvec_product(x, chunk):
    pixels = chunk['pixels']
    i = pixels['bin1_id']
    j = pixels['bin2_id']
    v = pixels['count'].astype(float)
    y = np.zeros(len(chunk['bins']['chrom']))
    _matvec_sparse_symmetric(y, i, j, v, x)
    return y


class MatVec(object):
    def __init__(self, clr, chunksize, map):
        self.clr = clr
        self.chunksize = chunksize
        self.map = map

    def __call__(self, x, mask):
        n = len(mask)

        x_full = np.zeros(n)
        x_full[mask] = x

        y_full = (
            split(self.clr, map=self.map, chunksize=self.chunksize)
            .pipe(_matvec_product, x_full)
            .reduce(add, np.zeros(n))
        )
        y = y_full[mask]

        return y



from itertools import chain
from collections import defaultdict
from functools import partial

import warnings

import numpy as np
import pandas as pd
from scipy.signal import fftconvolve
from scipy.interpolate import interp1d

from cooler.tools import partition
import cooler
import bioframe
from .lib import assign_supports, numutils

where = np.flatnonzero
concat = chain.from_iterable


def _contact_areas(distbins, scaffold_length):
    distbins = distbins.astype(float)
    scaffold_length = float(scaffold_length)
    outer_areas = np.maximum(scaffold_length - distbins[:-1], 0) ** 2
    inner_areas = np.maximum(scaffold_length - distbins[1:], 0) ** 2
    return 0.5 * (outer_areas - inner_areas)


def contact_areas(distbins, region1, region2):
    if region1 == region2:
        start, end = region1
        areas = _contact_areas(distbins, end - start)
    else:
        start1, end1 = region1
        start2, end2 = region2
        if start2 <= start1:
            start1, start2 = start2, start1
            end1, end2 = end2, end1
        areas = (
            _contact_areas(distbins, end2 - start1)
            - _contact_areas(distbins, start2 - start1)
            - _contact_areas(distbins, end2 - end1)
        )
        if end1 < start2:
            areas += _contact_areas(distbins, start2 - end1)

    return areas


def compute_scaling(df, region1, region2=None, dmin=int(1e1), dmax=int(1e7), n_bins=50):

    import dask.array as da

    if region2 is None:
        region2 = region1

    distbins = numutils.logbins(dmin, dmax, N=n_bins)
    areas = contact_areas(distbins, region1, region2)

    df = df[
        (df["pos1"] >= region1[0])
        & (df["pos1"] < region1[1])
        & (df["pos2"] >= region2[0])
        & (df["pos2"] < region2[1])
    ]
    dists = (df["pos2"] - df["pos1"]).values

    if isinstance(dists, da.Array):
        obs, _ = da.histogram(dists[(dists >= dmin) & (dists < dmax)], bins=distbins)
    else:
        obs, _ = np.histogram(dists[(dists >= dmin) & (dists < dmax)], bins=distbins)

    return distbins, obs, areas


def lattice_pdist_frequencies(n, points):
    """
    Distribution of pairwise 1D distances among a collection of distinct
    integers ranging from 0 to n-1.

    Parameters
    ----------
    n : int
        Size of the lattice on which the integer points reside.
    points : sequence of int
        Arbitrary integers between 0 and n-1, inclusive, in any order but
        with no duplicates.

    Returns
    -------
    h : 1D array of length n
        h[d] counts the number of integer pairs that are exactly d units apart

    Notes
    -----
    This is done using a convolution via FFT. Thanks to Peter de Rivaz; see
    `<http://stackoverflow.com/questions/42423823/distribution-of-pairwise-distances-between-many-integers>`_.

    """
    if len(np.unique(points)) != len(points):
        raise ValueError("Integers must be distinct.")
    x = np.zeros(n)
    x[points] = 1
    return np.round(fftconvolve(x, x[::-1], mode="full")).astype(int)[-n:]


def count_bad_pixels_per_diag(n, bad_bins):
    """
    Efficiently count the number of bad pixels on each upper diagonal of a
    matrix assuming a sequence of bad bins forms a "grid" of invalid pixels.

    Each bad bin bifurcates into two a row and column of bad pixels, so an
    upper bound on number of bad pixels per diagonal is 2*k, where k is the
    number of bad bins. For a given diagonal, we need to subtract from this
    upper estimate the contribution from rows/columns reaching "out-of-bounds"
    and the contribution of the intersection points of bad rows with bad
    columns that get double counted.

    ::

        o : bad bin
        * : bad pixel
        x : intersection bad pixel
        $ : out of bounds bad pixel
             $    $     $
         *--------------------------+
          *  *    *     *           |
           * *    *     *           |
            **    *     *           |
             o****x*****x***********|$
              *   *     *           |
               *  *     *           |
                * *     *           |
                 o******x***********|$
                  *     *           |
                   *    *           |
                    *   *           |
                     *  *           |
                      * *           |
                       **           |
                        o***********|$
                         *          |
                          *         |

    Parameters
    ----------
    n : int
        total number of bins
    bad_bins : 1D array of int
        sorted array of bad bin indexes

    Returns
    -------
    dcount : 1D array of length n
        dcount[d] == number of bad pixels on diagonal d

    """
    k = len(bad_bins)
    dcount = np.zeros(n, dtype=int)

    # Store all intersection pixels in a separate array
    # ~O(n log n) with fft
    ixn = lattice_pdist_frequencies(n, bad_bins)
    dcount[0] = ixn[0]

    # Keep track of out-of-bounds pixels by squeezing left and right bounds
    # ~O(n)
    pl = 0
    pr = k
    for diag in range(1, n):
        if pl < k:
            while (bad_bins[pl] - diag) < 0:
                pl += 1
                if pl == k:
                    break
        if pr > 0:
            while (bad_bins[pr - 1] + diag) >= n:
                pr -= 1
                if pr == 0:
                    break
        dcount[diag] = 2 * k - ixn[diag] - pl - (k - pr)
    return dcount


def count_all_pixels_per_diag(n):
    """
    Total number of pixels on each upper diagonal of a square matrix.

    Parameters
    ----------
    n : int
        total number of bins (dimension of square matrix)

    Returns
    -------
    dcount : 1D array of length n
        dcount[d] == total number of pixels on diagonal d

    """
    return np.arange(n, 0, -1)


def count_all_pixels_per_block(x, y):
    """
    Calculate total number of pixels in a rectangular block

    Parameters
    ----------
    x : int
        block width in pixels
    y : int
        block height in pixels

    Returns
    -------
    number_of_pixels : int
        total number of pixels in a block
    """
    return x * y


def count_bad_pixels_per_block(x, y, bad_bins_x, bad_bins_y):
    """
    Calculate number of "bad" pixels per rectangular block of a contact map

    "Bad" pixels are inferred from the balancing weight column `weight_name` or
    provided directly in the form of an array `bad_bins`.

    Setting `weight_name` and `bad_bins` to `None` yields 0 bad pixels in a block.

    Parameters
    ----------
    x : int
        block width in pixels
    y : int
        block height in pixels
    bad_bins_x : int
        number of bad bins on x-side
    bad_bins_y : int
        number of bad bins on y-side

    Returns
    -------
    number_of_pixes : int
        number of "bad" pixels in a block
    """

    # Calculate the resulting bad pixels in a rectangular block:
    return (x * bad_bins_y) + (y * bad_bins_x) - (bad_bins_x * bad_bins_y)


def make_diag_table(bad_mask, span1, span2):
    """
    Compute the total number of elements ``n_elem`` and the number of bad
    elements ``n_bad`` per diagonal for a single contact area encompassing
    ``span1`` and ``span2`` on the same genomic scaffold (cis matrix).

    Follows the same principle as the algorithm for finding contact areas for
    computing scalings.

    Parameters
    ----------
    bad_mask : 1D array of bool
        Mask of bad bins for the whole genomic scaffold containing the regions
        of interest.
    span1, span2 : pair of ints
        The bin spans (not genomic coordinates) of the two regions of interest.

    Returns
    -------
    diags : pandas.DataFrame
        Table indexed by 'diag' with columns ['n_elem', 'n_bad'].

    """

    def _make_diag_table(n_bins, bad_locs):
        diags = pd.DataFrame(index=pd.Series(np.arange(n_bins), name="diag"))
        diags["n_elem"] = count_all_pixels_per_diag(n_bins)
        diags["n_valid"] = diags["n_elem"] - count_bad_pixels_per_diag(n_bins, bad_locs)
        return diags

    if span1 == span2:
        lo, hi = span1
        diags = _make_diag_table(hi - lo, where(bad_mask[lo:hi]))
    else:
        lo1, hi1 = span1
        lo2, hi2 = span2
        if lo2 <= lo1:
            lo1, lo2 = lo2, lo1
            hi1, hi2 = hi2, hi1
        diags = (
            _make_diag_table(hi2 - lo1, where(bad_mask[lo1:hi2]))
            .subtract(
                _make_diag_table(lo2 - lo1, where(bad_mask[lo1:lo2])), fill_value=0
            )
            .subtract(
                _make_diag_table(hi2 - hi1, where(bad_mask[hi1:hi2])), fill_value=0
            )
        )
        if hi1 < lo2:
            diags.add(
                _make_diag_table(lo2 - hi1, where(bad_mask[hi1:lo2])), fill_value=0
            )
        diags = diags[diags["n_elem"] > 0]

    diags = diags.drop("n_elem", axis=1)
    return diags.astype(int)


def make_diag_tables(clr, regions, regions2=None, weight_name="weight", bad_bins=None):
    """
    For every support region infer diagonals that intersect this region
    and calculate the size of these intersections in pixels, both "total" and
    "n_valid", where "n_valid" does not include "bad" bins into counting.

    "Bad" pixels are inferred from the balancing weight column `weight_name` or
    provided directly in the form of an array `bad_bins`.

    Setting `weight_name` and `bad_bins` to `None` yields 0 "bad" pixels per
    diagonal per support region.

    When `regions2` are provided, all intersecting diagonals are reported for
    each rectangular and asymmetric block defined by combinations of matching
    elements of `regions` and `regions2`.
    Otherwise only `regions`-based symmetric square blocks are considered.
    Only intra-chromosomal regions are supported.

    Parameters
    ----------
    clr : cooler.Cooler
        Input cooler
    regions : list
        a list of genomic support regions
    regions2 : list
        a list of genomic support regions for asymmetric regions
    weight_name : str
        name of the weight vector in the "bins" table,
        if weight_name is None returns 0 for each block.
        Balancing weight are used to infer bad bins.
    bad_bins : array-like
        a list of bins to ignore. Indexes of bins must
        be absolute, as in clr.bins()[:], as opposed to
        being offset by chromosome start.
        "bad_bins" will be combined with the bad bins
        masked by balancing if there are any.

    Returns
    -------
    diag_tables : dict
        dictionary with DataFrames of relevant diagonals for every support.
    """

    regions = bioframe.parse_regions(regions, clr.chromsizes).values
    if regions2 is not None:
        regions2 = bioframe.parse_regions(regions2, clr.chromsizes).values

    bins = clr.bins()[:]
    if weight_name is None:
        # ignore bad bins
        sizes = dict(bins.groupby("chrom").size())
        bad_bin_dict = {
            chrom: np.zeros(sizes[chrom], dtype=bool) for chrom in sizes.keys()
        }
    elif isinstance(weight_name, str):
        # using balacning weight to infer bad bins
        if weight_name not in clr.bins().columns:
            raise KeyError("Balancing weight {weight_name} not found!")
        groups = dict(iter(bins.groupby("chrom")[weight_name]))
        bad_bin_dict = {
            chrom: np.array(groups[chrom].isnull()) for chrom in groups.keys()
        }
    else:
        raise ValueError("`weight_name` can be `str` or `None`")

    # combine custom "bad_bins" with "bad_bin_dict":
    if bad_bins is not None:
        # check if "bad_bins" are legit:
        try:
            bad_bins_chrom = bins.iloc[bad_bins].reset_index(drop=False)
        except IndexError:
            raise ValueError("Provided `bad_bins` are incorrect or out-of-bound")
        # group them by observed chromosomes only
        bad_bins_grp = bad_bins_chrom[["index", "chrom"]].groupby(
            "chrom", observed=True
        )
        # update "bad_bin_dict" with "bad_bins" for each chrom:
        for chrom, bin_ids in bad_bins_grp["index"]:
            co = clr.offset(chrom)
            # adjust by chromosome offset
            bad_bin_dict[chrom][bin_ids.values - co] = True

    diag_tables = {}
    for i in range(len(regions)):
        chrom, start1, end1, name1 = regions[i]
        if regions2 is not None:
            chrom2, start2, end2, name2 = regions2[i]
            # cis-only for now:
            assert chrom2 == chrom
        else:
            start2, end2 = start1, end1

        # translate regions into relative bin id-s:
        lo1, hi1 = clr.extent((chrom, start1, end1))
        lo2, hi2 = clr.extent((chrom, start2, end2))
        co = clr.offset(chrom)
        lo1 -= co
        lo2 -= co
        hi1 -= co
        hi2 -= co

        bad_mask = bad_bin_dict[chrom]
        newname = name1
        if regions2 is not None:
            newname = (name1, name2)
        diag_tables[newname] = make_diag_table(bad_mask, [lo1, hi1], [lo2, hi2])

    return diag_tables


def make_block_table(clr, regions1, regions2, weight_name="weight", bad_bins=None):
    """
    Creates a table that characterizes a set of rectangular genomic blocks
    formed by combining regions from regions1 and regions2.
    For every block calculate its "area" in pixels ("n_total"), and calculate
    number of "valid" pixels in each block ("n_valid").
    "Valid" pixels exclude "bad" pixels, which in turn inferred from the balancing
    weight column `weight_name` or provided directly in the form of an array of
    `bad_bins`.

    Setting `weight_name` and `bad_bins` to `None` yields 0 "bad" pixels per
    block.

    Parameters
    ----------
    clr : cooler.Cooler
        Input cooler
    regions1 : iterable
        a collection of genomic regions
    regions2 : iterable
        a collection of genomic regions
    weight_name : str
        name of the weight vector in the "bins" table,
        if weight_name is None returns 0 for each block.
        Balancing weight are used to infer bad bins.
    bad_bins : array-like
        a list of bins to ignore. Indexes of bins must
        be absolute, as in clr.bins()[:], as opposed to
        being offset by chromosome start.
        "bad_bins" will be combined with the bad bins
        masked by balancing if there are any.

    Returns
    -------
    block_table : dict
        dictionary for blocks that are 0-indexed
    """
    if bad_bins is None:
        bad_bins = np.asarray([]).astype(int)
    else:
        bad_bins = np.asarray(bad_bins).astype(int)

    regions1 = bioframe.parse_regions(regions1, clr.chromsizes).values
    regions2 = bioframe.parse_regions(regions2, clr.chromsizes).values

    # should we check for nestedness here, or that each region1 is < region2 ?

    block_table = {}
    for r1, r2 in zip(regions1, regions2):
        chrom1, start1, end1, name1 = r1
        chrom2, start2, end2, name2 = r2
        # translate regions into relative bin id-s:
        lo1, hi1 = clr.extent((chrom1, start1, end1))
        lo2, hi2 = clr.extent((chrom2, start2, end2))
        # width and height of a block:
        x = hi1 - lo1
        y = hi2 - lo2
        # get "regional" bad_bins for each of the regions
        bx = bad_bins[(bad_bins >= lo1) & (bad_bins < hi1)] - lo1
        by = bad_bins[(bad_bins >= lo2) & (bad_bins < hi2)] - lo2

        # now we need to combine it with the balancing weights
        if weight_name is None:
            bad_bins_x = len(bx)
            bad_bins_y = len(by)
        elif isinstance(weight_name, str):
            if weight_name not in clr.bins().columns:
                raise KeyError("Balancing weight {weight_name} not found!")
            else:
                # extract "bad" bins filtered by balancing:
                cb_bins_x = clr.bins()[weight_name][lo1:hi1].isnull().values
                cb_bins_y = clr.bins()[weight_name][lo2:hi2].isnull().values
                # combine with "bad_bins" using assignment:
                cb_bins_x[bx] = True
                cb_bins_y[by] = True
                # count and yield final list of bad bins:
                bad_bins_x = np.count_nonzero(cb_bins_x)
                bad_bins_y = np.count_nonzero(cb_bins_y)
        else:
            raise ValueError("`weight_name` can be `str` or `None`")

        # calculate total and bad pixels per block:
        n_tot = count_all_pixels_per_block(x, y)
        n_bad = count_bad_pixels_per_block(x, y, bad_bins_x, bad_bins_y)

        # fill in "block_table" with number of valid pixels:
        block_table[name1, name2] = defaultdict(int)
        block_table[name1, name2]["n_valid"] = n_tot - n_bad

    return block_table


def _diagsum_symm(clr, fields, transforms, regions, span):
    """
    calculates diagonal summary for a collection of
    square symmteric regions defined by regions.
    returns a dictionary of DataFrames with diagonal
    sums as values, and 0-based indexes of square
    genomic regions as keys.
    """
    lo, hi = span
    bins = clr.bins()[:]
    pixels = clr.pixels()[lo:hi]
    pixels = cooler.annotate(pixels, bins, replace=False)

    # this could further expanded to allow for custom groupings:
    pixels["dist"] = pixels["bin2_id"] - pixels["bin1_id"]
    for field, t in transforms.items():
        pixels[field] = t(pixels)

    diag_sums = {}
    # r define square symmetric block i:
    for i, r in enumerate(regions):
        r1 = assign_supports(pixels, [r], suffix="1")
        r2 = assign_supports(pixels, [r], suffix="2")
        # calculate diag_sums on the spot to allow for overlapping blocks:
        diag_sums[i] = pixels[(r1 == r2)].groupby("dist")[fields].sum()

    return diag_sums


def diagsum(
    clr,
    regions,
    transforms={},
    weight_name="weight",
    bad_bins=None,
    chunksize=10000000,
    ignore_diags=2,
    map=map,
):
    """

    Intra-chromosomal diagonal summary statistics.

    Parameters
    ----------
    clr : cooler.Cooler
        Cooler object
    regions : sequence of genomic range tuples
        Support regions for intra-chromosomal diagonal summation
    transforms : dict of str -> callable, optional
        Transformations to apply to pixels. The result will be assigned to
        a temporary column with the name given by the key. Callables take
        one argument: the current chunk of the (annotated) pixel dataframe.
    weight_name : str
        name of the balancing weight vector used to count
        "bad"(masked) pixels per diagonal.
        Use `None` to avoid masking "bad" pixels.
    bad_bins : array-like
        a list of bins to ignore per support region.
        Combines with the list of bad bins from balacning
        weight.
    chunksize : int, optional
        Size of pixel table chunks to process
    ignore_diags : int, optional
        Number of intial diagonals to exclude from statistics
    map : callable, optional
        Map functor implementation.

    Returns
    -------
    Dataframe of diagonal statistics for all regions

    """
    spans = partition(0, len(clr.pixels()), chunksize)
    fields = ["count"] + list(transforms.keys())

    regions = bioframe.parse_regions(regions, clr.chromsizes)

    dtables = make_diag_tables(clr, regions, weight_name=weight_name, bad_bins=bad_bins)

    # combine masking with existing transforms and add a "count" transform:
    if bad_bins is not None:
        # turn bad_bins into a mask of size clr.bins:
        mask_size = len(clr.bins())
        bad_bins_mask = np.ones(mask_size, dtype=int)
        bad_bins_mask[bad_bins] = 0
        #
        masked_transforms = {}
        bin1 = "bin1_id"
        bin2 = "bin2_id"
        for field in fields:
            if field in transforms:
                # combine masking and transform, minding the scope:
                t = transforms[field]
                masked_transforms[field] = (
                    lambda p, t=t, m=bad_bins_mask: t(p) * m[p[bin1]] * m[p[bin2]]
                )
            else:
                # presumably field == "count", mind the scope as well:
                masked_transforms[field] = (
                    lambda p, f=field, m=bad_bins_mask: p[f] * m[p[bin1]] * m[p[bin2]]
                )
        # substitute transforms to the masked_transforms:
        transforms = masked_transforms

    for dt in dtables.values():
        for field in fields:
            agg_name = "{}.sum".format(field)
            dt[agg_name] = 0

    job = partial(_diagsum_symm, clr, fields, transforms, regions.values)
    results = map(job, spans)
    for result in results:
        for i, agg in result.items():
            region = regions.loc[i, "name"]
            for field in fields:
                agg_name = "{}.sum".format(field)
                dtables[region][agg_name] = dtables[region][agg_name].add(
                    agg[field], fill_value=0
                )

    if ignore_diags:
        for dt in dtables.values():
            for field in fields:
                agg_name = "{}.sum".format(field)
                j = dt.columns.get_loc(agg_name)
                dt.iloc[:ignore_diags, j] = np.nan

    # returning dataframe for API consistency
    result = []
    for i, dtable in dtables.items():
        dtable = dtable.reset_index()
        dtable.insert(0, "region", i)
        result.append(dtable)
    return pd.concat(result).reset_index(drop=True)


def _diagsum_asymm(clr, fields, transforms, regions1, regions2, span):
    """
    calculates diagonal summary for a collection of
    rectangular regions defined as combinations of
    regions1 and regions2.
    returns a dictionary of DataFrames with diagonal
    sums as values, and 0-based indexes of rectangular
    genomic regions as keys.
    """
    lo, hi = span
    bins = clr.bins()[:]
    pixels = clr.pixels()[lo:hi]
    pixels = cooler.annotate(pixels, bins, replace=False)

    # this could further expanded to allow for custom groupings:
    pixels["dist"] = pixels["bin2_id"] - pixels["bin1_id"]
    for field, t in transforms.items():
        pixels[field] = t(pixels)

    diag_sums = {}
    # r1 and r2 define rectangular block i:
    for i, (r1, r2) in enumerate(zip(regions1, regions2)):
        r1 = assign_supports(pixels, [r1], suffix="1")
        r2 = assign_supports(pixels, [r2], suffix="2")
        # calculate diag_sums on the spot to allow for overlapping blocks:
        diag_sums[i] = pixels[(r1 == r2)].groupby("dist")[fields].sum()

    return diag_sums


def diagsum_asymm(
    clr,
    regions1,
    regions2,
    transforms={},
    weight_name="weight",
    bad_bins=None,
    chunksize=10000000,
    map=map,
):
    """

    Diagonal summary statistics.

    Matchings elements of `regions1` and  `regions2` define
    asymmetric rectangular blocks for calculating diagonal
    summary statistics.
    Only intra-chromosomal blocks are supported.

    Parameters
    ----------
    clr : cooler.Cooler
        Cooler object
    regions1 : sequence of genomic range tuples
        "left"-side support regions for diagonal summation
    regions2 : sequence of genomic range tuples
        "right"-side support regions for diagonal summation
    transforms : dict of str -> callable, optional
        Transformations to apply to pixels. The result will be assigned to
        a temporary column with the name given by the key. Callables take
        one argument: the current chunk of the (annotated) pixel dataframe.
    weight_name : str
        name of the balancing weight vector used to count
        "bad"(masked) pixels per diagonal.
        Use `None` to avoid masking "bad" pixels.
    bad_bins : array-like
        a list of bins to ignore per support region.
        Combines with the list of bad bins from balacning
        weight.
    chunksize : int, optional
        Size of pixel table chunks to process
    map : callable, optional
        Map functor implementation.

    Returns
    -------
    DataFrame with summary statistic of every diagonal of every block:
    region1, region2, diag, n_valid, count.sum

    """
    spans = partition(0, len(clr.pixels()), chunksize)
    fields = ["count"] + list(transforms.keys())
    regions1 = bioframe.parse_regions(regions1, clr.chromsizes)
    regions2 = bioframe.parse_regions(regions2, clr.chromsizes)

    dtables = make_diag_tables(
        clr, regions1, regions2, weight_name=weight_name, bad_bins=bad_bins
    )

    # combine masking with existing transforms and add a "count" transform:
    if bad_bins is not None:
        # turn bad_bins into a mask of size clr.bins:
        mask_size = len(clr.bins())
        bad_bins_mask = np.ones(mask_size, dtype=int)
        bad_bins_mask[bad_bins] = 0
        #
        masked_transforms = {}
        bin1 = "bin1_id"
        bin2 = "bin2_id"
        for field in fields:
            if field in transforms:
                # combine masking and transform, minding the scope:
                t = transforms[field]
                masked_transforms[field] = (
                    lambda p, t=t, m=bad_bins_mask: t(p) * m[p[bin1]] * m[p[bin2]]
                )
            else:
                # presumably field == "count", mind the scope as well:
                masked_transforms[field] = (
                    lambda p, f=field, m=bad_bins_mask: p[f] * m[p[bin1]] * m[p[bin2]]
                )
        # substitute transforms to the masked_transforms:
        transforms = masked_transforms

    for dt in dtables.values():
        for field in fields:
            agg_name = "{}.sum".format(field)
            dt[agg_name] = 0

    job = partial(
        _diagsum_asymm, clr, fields, transforms, regions1.values, regions2.values
    )
    results = map(job, spans)
    for result in results:
        for i, agg in result.items():
            region1 = regions1.loc[i, "name"]
            region2 = regions2.loc[i, "name"]
            for field in fields:
                agg_name = "{}.sum".format(field)
                dtables[region1, region2][agg_name] = dtables[region1, region2][
                    agg_name
                ].add(agg[field], fill_value=0)

    # returning a dataframe for API consistency:
    result = []
    for (i, j), dtable in dtables.items():
        dtable = dtable.reset_index()
        dtable.insert(0, "region1", i)
        dtable.insert(1, "region2", j)
        result.append(dtable)
    return pd.concat(result).reset_index(drop=True)


def _blocksum_asymm(clr, fields, transforms, regions1, regions2, span):
    """
    calculates block summary for a collection of
    rectangular regions defined as combinations of
    regions1 and regions2.
    returns a dictionary of with block sums as values,
    and 0-based indexes of rectangular genomic regions
    as keys.
    """
    lo, hi = span
    bins = clr.bins()[:]
    pixels = clr.pixels()[lo:hi]
    pixels = cooler.annotate(pixels, bins, replace=False)

    for field, t in transforms.items():
        pixels[field] = t(pixels)

    block_sums = {}
    # r1 and r2 define rectangular block i:
    for i, (r1, r2) in enumerate(zip(regions1, regions2)):
        r1 = assign_supports(pixels, [r1], suffix="1")
        r2 = assign_supports(pixels, [r2], suffix="2")
        # calculate sum on the spot to allow for overlapping blocks:
        block_sums[i] = pixels[(r1 == r2)][fields].sum()

    return block_sums


def blocksum_asymm(
    clr,
    regions1,
    regions2,
    transforms={},
    weight_name="weight",
    bad_bins=None,
    chunksize=1000000,
    map=map,
):
    """
    Summary statistics on rectangular blocks of genomic regions.

    Parameters
    ----------
    clr : cooler.Cooler
        Cooler object
    regions1 : sequence of genomic range tuples
        "left"-side support regions for diagonal summation
    regions2 : sequence of genomic range tuples
        "right"-side support regions for diagonal summation
    transforms : dict of str -> callable, optional
        Transformations to apply to pixels. The result will be assigned to
        a temporary column with the name given by the key. Callables take
        one argument: the current chunk of the (annotated) pixel dataframe.
    weight_name : str
        name of the balancing weight vector used to count
        "bad"(masked) pixels per block.
        Use `None` to avoid masking "bad" pixels.
    bad_bins : array-like
        a list of bins to ignore per support region.
        Combines with the list of bad bins from balacning
        weight.
    chunksize : int, optional
        Size of pixel table chunks to process
    map : callable, optional
        Map functor implementation.

    Returns
    -------
    DataFrame with entries for each blocks: region1, region2, n_valid, count.sum

    """

    regions1 = bioframe.parse_regions(regions1, clr.chromsizes)
    regions2 = bioframe.parse_regions(regions2, clr.chromsizes)

    spans = partition(0, len(clr.pixels()), chunksize)
    fields = ["count"] + list(transforms.keys())

    # similar with diagonal summations, pre-generate a block_table listing
    # all of the rectangular blocks and "n_valid" number of pixels per each block:
    records = make_block_table(
        clr, regions1, regions2, weight_name=weight_name, bad_bins=bad_bins
    )

    # combine masking with existing transforms and add a "count" transform:
    if bad_bins is not None:
        # turn bad_bins into a mask of size clr.bins:
        mask_size = len(clr.bins())
        bad_bins_mask = np.ones(mask_size, dtype=int)
        bad_bins_mask[bad_bins] = 0
        #
        masked_transforms = {}
        bin1 = "bin1_id"
        bin2 = "bin2_id"
        for field in fields:
            if field in transforms:
                # combine masking and transform, minding the scope:
                t = transforms[field]
                masked_transforms[field] = (
                    lambda p, t=t, m=bad_bins_mask: t(p) * m[p[bin1]] * m[p[bin2]]
                )
            else:
                # presumably field == "count", mind the scope as well:
                masked_transforms[field] = (
                    lambda p, f=field, m=bad_bins_mask: p[f] * m[p[bin1]] * m[p[bin2]]
                )
        # substitute transforms to the masked_transforms:
        transforms = masked_transforms

    job = partial(
        _blocksum_asymm, clr, fields, transforms, regions1.values, regions2.values
    )
    results = map(job, spans)
    for result in results:
        for i, agg in result.items():
            for field in fields:
                agg_name = "{}.sum".format(field)
                s = agg[field].item()
                if not np.isnan(s):
                    n1 = regions1.loc[i, "name"]
                    n2 = regions2.loc[i, "name"]
                    records[n1, n2][agg_name] += s

    # returning a dataframe for API consistency:
    return pd.DataFrame(
        [{"region1": n1, "region2": n2, **rec} for (n1, n2), rec in records.items()],
        columns=["region1", "region2", "n_valid", "count.sum"]
        + [k + ".sum" for k in transforms.keys()],
    )


def diagsum_from_array(
    A, counts=None, *, offset=0, ignore_diags=2, filter_counts=False, region_name=None
):
    """
    Calculates Open2C-formatted expected for a dense submatrix of a whole
    genome contact map.

    Parameters
    ----------
    A : 2D array
        Normalized submatrix to calculate expected (``balanced.sum``).
    counts : 2D array or None, optional
        Corresponding raw contacts to populate ``count.sum``.
    offset : int or (int, int)
        i- and j- bin offsets of A relative to the parent matrix. If a single
        offset is provided it is applied to both axes.
    ignore_diags : int, optional
        Number of initial diagonals to ignore.
    filter_counts : bool, optional
        Apply the validity mask from balanced matrix to the raw one. Ignored
        when counts is None.
    region_name : str or (str, str), optional
        A custom region name or pair of region names. If provided, region
        columns will be included in the output.

    Notes
    -----
    For regions that cross the main diagonal of the whole-genome contact map,
    the lower triangle "overhang" is ignored.

    Examples
    --------
    >>> A = clr.matrix()[:, :]  # whole genome balanced
    >>> C = clr.matrix(balance=False)[:, :]  # whole genome raw

    Using only balanced data:
    >>> exp = diagsum_from_array(A)

    Using balanced and raw counts:
    >>> exp1 = diagsum_from_array(A, C)

    Using an off-diagonal submatrix
    >>> exp2 = diagsum_from_array(A[:50, 50:], offset=(0, 50))

    """
    if isinstance(offset, (list, tuple)):
        offset1, offset2 = offset
    else:
        offset1, offset2 = offset, offset
    if isinstance(region_name, (list, tuple)):
        region1, region2 = region_name
    elif region_name is not None:
        region1, region2 = region_name, region_name
    A = np.asarray(A, dtype=float)
    if counts is not None:
        counts = np.asarray(counts)
        if counts.shape != A.shape:
            raise ValueError("`counts` must have the same shape as `A`.")

    # Compute validity mask for bins on each axis
    A[~np.isfinite(A)] = 0
    invalid_mask1 = np.sum(A, axis=1) == 0
    invalid_mask2 = np.sum(A, axis=0) == 0

    # Prepare an indicator matrix of "diagonals" (toeplitz) where the lower
    # triangle diagonals wrt the parent matrix are negative.
    # The "outer difference" operation below produces a toeplitz matrix.
    lo1, hi1 = offset1, offset1 + A.shape[0]
    lo2, hi2 = offset2, offset2 + A.shape[1]
    ar1 = np.arange(lo1, hi1, dtype=np.int32)
    ar2 = np.arange(lo2, hi2, dtype=np.int32)
    diag_indicator = ar2[np.newaxis, :] - ar1[:, np.newaxis]
    diag_lo = max(lo2 - hi1 + 1, 0)
    diag_hi = hi2 - lo1

    # Apply the validity mask to the indicator matrix.
    # Both invalid and lower triangle pixels will now have negative indicator values.
    D = diag_indicator.copy()
    D[invalid_mask1, :] = -1
    D[:, invalid_mask2] = -1
    # Drop invalid and lower triangle pixels and flatten.
    mask_per_pixel = D >= 0
    A_flat = A[mask_per_pixel]
    D_flat = D[mask_per_pixel]

    # Group by diagonal and aggregate the number of valid pixels and pixel values.
    diagonals = np.arange(diag_lo, diag_hi, dtype=int)
    n_valid = np.bincount(D_flat, minlength=diag_hi - diag_hi)[diag_lo:]
    balanced_sum = np.bincount(D_flat, weights=A_flat, minlength=diag_hi - diag_lo)[
        diag_lo:
    ]
    # Mask to ignore initial diagonals.
    mask_per_diag = diagonals >= ignore_diags

    # Populate the output dataframe.
    # Include region columns if region names are provided.
    # Include raw pixel counts for each diag if counts is provided.
    df = pd.DataFrame({"diag": diagonals, "n_valid": n_valid})

    if region_name is not None:
        df.insert(0, "region1", region1)
        df.insert(1, "region2", region2)

    if counts is not None:
        # Either count everything or apply the same filtering as A.
        if filter_counts:
            C_flat = counts[mask_per_pixel]
            count_sum = np.bincount(
                D_flat, weights=C_flat, minlength=diag_hi - diag_lo
            )[diag_lo:]
        else:
            mask_per_pixel = diag_indicator >= 0
            D_flat = diag_indicator[mask_per_pixel]
            C_flat = counts[mask_per_pixel]
            count_sum = np.bincount(
                D_flat, weights=C_flat, minlength=diag_hi - diag_lo
            )[diag_lo:]
        count_sum[~mask_per_diag] = np.nan
        df["count.sum"] = count_sum

    balanced_sum[~mask_per_diag] = np.nan
    df["balanced.sum"] = balanced_sum

    return df


def logbin_expected(
    exp,
    summary_name="balanced.sum",
    bins_per_order_magnitude=10,
    bin_layout="fixed",
    smooth=lambda x: numutils.robust_gauss_filter(x, 2),
    min_nvalid=200,
    min_count=50,
):
    """
    Logarithmically bins expected as produced by diagsum method.

    Parameters
    ----------
    exp : DataFrame
        DataFrame produced by diagsum

    summary_name : str, optional
        Name of the column of exp-DataFrame to use as a diagonal summary.
        Default is "balanced.sum".

    bins_per_order_magnitude : int, optional
        How many bins per order of magnitude. Default of 10 has a ratio of
        neighboring bins of about 1.25

    bin_layout : "fixed", "longest_region", or array
        "fixed" means that bins are exactly the same for different datasets,
        and only depend on bins_per_order_magnitude

        "longest_region" means that the last bin will end at size of the
        longest region.
            GOOD: the last bin will have as much data as possible.
            BAD: bin edges will end up different for different datasets, you
            can't divide them by each other

        array: provide your own bin edges. Can be of any size, and end at any
        value. Bins exceeding the size of the largest region will be simply
        ignored.

    smooth : callable
        A smoothing function to be applied to log(P(s)) and log(x)
        before calculating P(s) slopes for by-region data

    min_nvalid : int
        For each region, throw out bins (log-spaced) that have less than
        min_nvalid valid pixels
        This will ensure that each entree in Pc_by_region has at least n_valid
        valid pixels
        Don't set it to zero, or it will introduce bugs. Setting it to 1 is OK,
        but not recommended.

    min_count : int
        If counts are found in the data, then for each region, throw out bins
        (log-spaced)
        that have more than min_counts of counts.sum (raw Hi-C counts).
        This will ensure that each entree in Pc_by_region has at least
        min_count raw Hi-C reads

    Returns
    -------
    Pc : DataFrame
        dataframe of contact probabilities and spread across regions
    slope : ndarray
        slope of Pc(s) on a log-log plot and spread across regions
    bins : ndarray
        an array of bin edges used for calculating P(s)

    Notes
    -----
    For main Pc and slope, the algorithm is the following

    1. concatenate all the expected for all regions into a large dataframe.
    2. create logarithmically-spaced bins of diagonals (or use provided)
    3. pool together n_valid and balanced.sum for each region and for each bin
    4. calculate the average diagonal for each bucket, weighted by n_valid
    5. divide balanced.sum by n_valid after summing for each bucket (not before)
    6. calculate the slope in log space (for each region)

    X values are not midpoints of bins
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    In step 4, we calculate the average diag index weighted by n_valid. This
    seems counter-intuitive, but it actually is justified.

    Let's take the worst case scenario. Let there be a bin from 40MB to 44MB.
    Let there be a region that is exactly 41 MB long. The midpoint of the bin
    is at 42MB. But the only part of this region belonging to this bin is
    actually between 40MB and 41MB. Moreover, the "average" read in this
    little triangle of the heatmap is actually not coming even from 40.5 MB
    because the triangle is getting narrower towards 41MB. The center of mass
    of a triangle is 1/3 of the way up, or 40.33 MB. So an average read for
    this region in this bin is coming from 40.33.

    Consider the previous bin, say, from 36MB to 40MB. The heatmap there is a
    trapezoid with a long side of 5MB, the short side of 1MB, and height of
    4MB. The center of mass of this trapezoid is at 36 + 14/9 = 37.55MB,
    and not at 38MB. So the last bin center is definitely mis-assigned, and
    the second-to-last bin center is off by some 25%. This would lead to a 25%
    error of the P(s) slope estimated between the third-to-last and
    second-to-last bin.

    In presence of missing bins, this all becomes more complex, but this kind
    of averaging should take care of everything. It follows a general
    principle: when averaging the y values with some weights, one needs to
    average the x values with the same weights. The y values here are being
    added together, so per-diag means are effectively averaged with the weight
    of n_valid. Therefore, the x values (diag) should be averaged with the
    same weights.

    Other considerations
    ~~~~~~~~~~~~~~~~~~~~
    Steps #3 and #5 are important because the ratio of sums does not equal to
    the sum of ratios, and the former is more correct (the latter is more
    susceptible to noise). It is generally better to divide at the very end,
    rather than dividing things for each diagonal.

    Here we divide at the end twice: first we divide balanced.sum by n_valid
    for each region, then we effectively multiply it back up and divide it for
    each bin when combining different regions (see weighted average in the
    next function).

    Smoothing P(s) for the slope
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    For calcuating the slope, we apply smoothing to the P(s) to ensure the
    slope is not too noisy. There are several caveats here: the P(s) has to
    be smoothed in logspace, and both P and s have to be smoothed. It is
    discussed in detail here

    https://gist.github.com/mimakaev/4becf1310ba6ee07f6b91e511c531e73

    Examples
    --------
    For example, see this gist: https://gist.github.com/mimakaev/e9117a7fcc318e7904702eba5b47d9e6

    """
    from cooltools.lib.numutils import logbins

    raw_summary_name = "count.sum"
    exp_summary_base, *_ = summary_name.split(".")
    Pc_name = f"{exp_summary_base}.avg"
    diag_name = "diag"
    diag_avg_name = f"{diag_name}.avg"

    exp = exp[~pd.isna(exp[summary_name])].copy()
    exp[diag_avg_name] = exp.pop(diag_name)  # "average" or weighted diagonals
    diagmax = exp[diag_avg_name].max()

    # create diag_bins based on chosen layout:
    if bin_layout == "fixed":
        diag_bins = numutils.persistent_log_bins(
            10, bins_per_order_magnitude=bins_per_order_magnitude
        )
    elif bin_layout == "longest_region":
        diag_bins = logbins(1, diagmax + 1, ratio=10 ** (1 / bins_per_order_magnitude))
    else:
        diag_bins = bin_layout

    if diag_bins[-1] < diagmax:
        raise ValueError(
            "Genomic separation bins end is less than the size of the largest region"
        )

    # assign diagonals in exp DataFrame to diag_bins, i.e. give them ids:
    exp["diag_bin_id"] = (
        np.searchsorted(diag_bins, exp[diag_avg_name], side="right") - 1
    )
    exp = exp[exp["diag_bin_id"] >= 0]

    # constructing expected grouped by region
    byReg = exp.copy()

    # this averages diag_avg with the weight equal to n_valid, and sums everything else
    byReg[diag_avg_name] *= byReg["n_valid"]
    byRegExp = byReg.groupby(["region", "diag_bin_id"]).sum()
    byRegExp[diag_avg_name] /= byRegExp["n_valid"]

    byRegExp = byRegExp.reset_index()
    byRegExp = byRegExp[byRegExp["n_valid"] > min_nvalid]  # filtering by n_valid
    byRegExp[Pc_name] = byRegExp[summary_name] / byRegExp["n_valid"]
    byRegExp = byRegExp[byRegExp[Pc_name] > 0]  # drop diag_bins with 0 counts
    if min_count:
        if raw_summary_name in byRegExp:
            byRegExp = byRegExp[byRegExp[raw_summary_name] > min_count]
        else:
            warnings.warn(
                RuntimeWarning(f"{raw_summary_name} not found in the input expected")
            )

    byRegExp["diag_bin_start"] = diag_bins[byRegExp["diag_bin_id"].values]
    byRegExp["diag_bin_end"] = diag_bins[byRegExp["diag_bin_id"].values + 1] - 1

    byRegDer = []
    for reg, subdf in byRegExp.groupby("region"):
        subdf = subdf.sort_values("diag_bin_id")
        valid = np.minimum(subdf["n_valid"].values[:-1], subdf["n_valid"].values[1:])
        mids = np.sqrt(
            subdf[diag_avg_name].values[:-1] * subdf[diag_avg_name].values[1:]
        )
        slope = np.diff(smooth(np.log(subdf[Pc_name].values))) / np.diff(
            smooth(np.log(subdf[diag_avg_name].values))
        )
        newdf = pd.DataFrame(
            {
                diag_avg_name: mids,
                "slope": slope,
                "n_valid": valid,
                "diag_bin_id": subdf["diag_bin_id"].values[:-1],
            }
        )
        newdf["region"] = reg
        byRegDer.append(newdf)
    byRegDer = pd.concat(byRegDer).reset_index(drop=True)
    return byRegExp, byRegDer, diag_bins[: byRegExp["diag_bin_id"].max() + 2]


def combine_binned_expected(
    binned_exp,
    binned_exp_slope=None,
    Pc_name="balanced.avg",
    der_smooth_function_combined=lambda x: numutils.robust_gauss_filter(x, 1.3),
    spread_funcs="logstd",
    spread_funcs_slope="std",
    minmax_drop_bins=2,
    concat_original=False,
):
    """
    Combines by-region log-binned expected and slopes into genome-wide averages,
    handling small chromosomes and "corners" in an optimal fashion, robust to
    outliers. Calculates spread of by-chromosome P(s) and slopes, also in an optimal fashion.

    Parameters
    ----------
    binned_exp: dataframe
        binned expected as outputed by logbin_expected

    binned_exp_slope : dataframe or None
        If provided, estimates spread of slopes.
        Is necessary if concat_original is True

    Pc_name : str
        Name of the column with the probability of contacts.
        Defaults to "balanced.avg".

    der_smooth_function_combined : callable
        A smoothing function for calculating slopes on combined data

    spread_funcs: "minmax", "std", "logstd" or a function (see below)
        A way to estimate the spread of the P(s) curves between regions.
        * "minmax" - use the minimum/maximum of by-region P(s)
        * "std" - use weighted standard deviation of P(s) curves (may produce negative results)
        * "logstd" (recommended) weighted standard deviation in logspace (as seen on the plot)

    spread_funcs_slope: "minmax", "std" or a funciton
        Similar to spread_func, but for slopes rather than P(s)

    concat_original: bool (default = False)
        Append original dataframe, and put combined under region "combined"

    Returns
    -------
    scal, slope_df

    Notes
    -----
    This function does not calculate errorbars. The spread is not the deviation of the mean,
    and rather is representative of variability between chromosomes.


    Calculating errorbars/spread

    1. Take all by-region P(s)
    2. For "minmax", remove the last var_drop_last_bins bins for each region
       (by default two. They are most noisy and would inflate the
       spread for the last points). Min/max are most susceptible to this.
    3. Groupby P(s) by region
    4. Apply spread_funcs to the pd.GroupBy object. Options are:
       * minimum and maximum ("minmax"),
       * weighted standard deviation ("std"),
       * weighted standard deviation in logspace ("logstd", default) or two custom functions
       We do not remove the last bins for "std" / "logstd" because we are
       doing weighted standard deviation. Therefore, noisy "ends" of regions
       would contribute very little to this.
    5. Append them to the P(s) for the same bin.

    As a result, by for minmax, we do not estimate spread for the last
    two bins. This is because there are often very few chromosomal arms there,
    and different arm measurements are noisy. For other methods, we do
    estimate the spread there, and noisy last bins are taken care of by the
    weighted standard deviation. However, the spread in the last bins may be
    noisy, and may become a 0 if only one region is contributing to the last
    pixel.
    """
    diag_avg_name = "diag.avg"
    # combine pre-logbinned expecteds
    scal = numutils.weighted_groupby_mean(
        binned_exp[
            [
                Pc_name,
                "diag_bin_id",
                "n_valid",
                diag_avg_name,
                "diag_bin_start",
                "diag_bin_end",
            ]
        ],
        group_by="diag_bin_id",
        weigh_by="n_valid",
        mode="mean",
    )

    # for every diagonal calculate the spread of expected
    if spread_funcs == "minmax":
        byRegVar = binned_exp.copy()
        byRegVar = byRegVar.loc[
            byRegVar.index.difference(
                byRegVar.groupby("region")["n_valid"].tail(minmax_drop_bins).index
            )
        ]
        low_err = byRegVar.groupby("diag_bin_id")[Pc_name].min()
        high_err = byRegVar.groupby("diag_bin_id")[Pc_name].max()
    elif spread_funcs == "std":
        var = numutils.weighted_groupby_mean(
            binned_exp[[Pc_name, "diag_bin_id", "n_valid"]],
            group_by="diag_bin_id",
            weigh_by="n_valid",
            mode="std",
        )[Pc_name]
        low_err = scal[Pc_name] - var
        high_err = scal[Pc_name] + var
    elif spread_funcs == "logstd":
        var = numutils.weighted_groupby_mean(
            binned_exp[[Pc_name, "diag_bin_id", "n_valid"]],
            group_by="diag_bin_id",
            weigh_by="n_valid",
            mode="logstd",
        )[Pc_name]
        low_err = scal[Pc_name] / var
        high_err = scal[Pc_name] * var
    else:
        low_err, high_err = spread_funcs(binned_exp, scal)

    scal["low_err"] = low_err
    scal["high_err"] = high_err

    # re-calculate slope of the combined expected (log,smooth,diff)
    f = der_smooth_function_combined
    slope = np.diff(f(np.log(scal[Pc_name].values))) / np.diff(
        f(np.log(scal[diag_avg_name].values))
    )
    valid = np.minimum(scal["n_valid"].values[:-1], scal["n_valid"].values[1:])
    mids = np.sqrt(scal[diag_avg_name].values[:-1] * scal[diag_avg_name].values[1:])
    slope_df = pd.DataFrame(
        {
            diag_avg_name: mids,
            "slope": slope,
            "n_valid": valid,
            "diag_bin_id": scal.index.values[:-1],
        }
    )
    slope_df = slope_df.set_index("diag_bin_id")

    # when pre-region slopes are provided, calculate spread of slopes
    if binned_exp_slope is not None:
        if spread_funcs_slope == "minmax":
            byRegDer = binned_exp_slope.copy()
            byRegDer = byRegDer.loc[
                byRegDer.index.difference(
                    byRegDer.groupby("region")["n_valid"].tail(minmax_drop_bins).index
                )
            ]
            low_err = byRegDer.groupby("diag_bin_id")["slope"].min()
            high_err = byRegDer.groupby("diag_bin_id")["slope"].max()
        elif spread_funcs_slope == "std":
            var = numutils.weighted_groupby_mean(
                binned_exp_slope[["slope", "diag_bin_id", "n_valid"]],
                group_by="diag_bin_id",
                weigh_by="n_valid",
                mode="std",
            )["slope"]
            low_err = slope_df["slope"] - var
            high_err = slope_df["slope"] + var

        else:
            low_err, high_err = spread_funcs_slope(binned_exp_slope, scal)
        slope_df["low_err"] = low_err
        slope_df["high_err"] = high_err

    slope_df = slope_df.reset_index()
    scal = scal.reset_index()

    # append "combined" expected/slopes to the input DataFrames (not in-place)
    if concat_original:
        scal["region"] = "combined"
        slope_df["region"] = "combined"
        scal = pd.concat([scal, binned_exp], sort=False).reset_index(drop=True)
        slope_df = pd.concat([slope_df, binned_exp_slope], sort=False).reset_index(
            drop=True
        )

    return scal, slope_df


def interpolate_expected(
    expected,
    binned_expected,
    columns=["balanced.avg"],
    kind="quadratic",
    by_region=True,
    extrapolate_small_s=False,
):
    """
    Interpolates expected to match binned_expected.
    Basically, this function smoothes the original expected according to the logbinned expected.
    It could either use by-region expected (each region will have different expected)
    or use combined binned_expected (all regions will have the same expected after that)

    Such a smoothed expected should be used to calculate observed/expected for downstream analysis.

    Parameters
    ----------
    expected: pd.DataFrame
        expected as returned by diagsum
    binned_expected: pd.DataFrame
        binned expected (combined or not)
    columns: list[str] (optional)
        Columns to interpolate. Must be present in binned_expected,
        but not necessarily in expected.
    kind: str (optional)
        Interpolation type, according to scipy.interpolate.interp1d
    by_region: bool or str (optional)
        Whether to do interpolation by-region (default=True).
        False means use one expected for all regions (use entire table).
        If a region name is provided, expected for that region is used.

    """

    exp_int = expected.copy()
    gr_exp = exp_int.groupby("region")  # groupby original expected by region

    if by_region is not False and "region" not in binned_expected:
        warnings.warn("Region column not found, assuming combined expected")
        by_region = False

    if by_region is True:
        # groupby expected
        gr_binned = binned_expected.groupby("region")
    elif by_region is not False:
        # extract a region that we want to use
        binned_expected = binned_expected[binned_expected["region"] == by_region]

    if by_region is not True:
        # check that we have no duplicates in expected
        assert len(binned_expected["diag_bin_id"].drop_duplicates()) == len(
            binned_expected
        )

    interp_dfs = []

    for reg, df_orig in gr_exp:
        if by_region is True:  # use binned expected for this region
            if reg not in gr_binned.groups:
                continue
            subdf = gr_binned.get_group(reg)
        else:
            subdf = binned_expected

        diag_orig = df_orig["diag"].values
        diag_mid = (subdf["diag_bin_start"] + subdf["diag_bin_end"]) / 2
        interp_df = pd.DataFrame(
            index=df_orig.index
        )  # df to put interpolated values in
        with np.errstate(invalid="ignore", divide="ignore"):
            for colname in columns:  # interpolate each column
                value_column = subdf[colname]
                interp = interp1d(
                    np.log(diag_mid),
                    np.log(value_column),
                    kind=kind,
                    fill_value="extrapolate",
                )
                interp_df[colname] = np.exp(interp(np.log(diag_orig)))
            if not extrapolate_small_s:
                mask = diag_orig >= subdf["diag_bin_start"].min()
                interp_df = interp_df.iloc[mask]
        interp_dfs.append(interp_df)
    interp_df = pd.concat(interp_dfs)
    for i in interp_df.columns:
        exp_int[i] = interp_df[i]
    return exp_int



import numpy as np
import pandas as pd

import cooler
import cooler.tools


def sample_pixels_approx(pixels, frac):
    pixels["count"] = np.random.binomial(pixels["count"], frac)
    mask = pixels["count"] > 0

    if issubclass(type(pixels), pd.DataFrame):
        pixels = pixels[mask]
    elif issubclass(type(pixels), dict):
        pixels = {k: arr[mask] for k, arr in pixels.items()}
    return pixels


def sample_pixels_exact(pixels, count):
    cumcount = np.cumsum(np.asarray(pixels["count"]))
    total = cumcount[-1]
    n_pixels = cumcount.shape[0]

    # sample a given number of distinct contacts
    random_contacts = np.random.choice(total, size=count, replace=False)

    # find where those contacts live in the cumcount array
    loc = np.searchsorted(cumcount, random_contacts, side="right")

    # re-bin those locations to get new counts
    new_counts = np.bincount(loc, minlength=n_pixels)

    pixels["count"] = new_counts
    mask = pixels["count"] > 0
    if issubclass(type(pixels), pd.DataFrame):
        pixels = pixels[mask]
    elif issubclass(type(pixels), dict):
        pixels = {k: arr[mask] for k, arr in pixels.items()}
    return pixels


def _extract_pixel_chunk(chunk):
    return chunk["pixels"]


def sample_cooler(
    clr,
    out_clr_path,
    count=None,
    frac=None,
    exact=False,
    map_func=map,
    chunksize=int(1e7),
):
    """
    Pick a random subset of contacts from a Hi-C map.

    Parameters
    ----------
    clr : cooler.Cooler or str
        A Cooler or a path/URI to a Cooler with input data.

    out_clr_path : str
        A path/URI to the output.

    count : float
        The target number of contacts in the sample.
        Mutually exclusive with `frac`.

    frac : float
        The target sample size as a fraction of contacts in the original
        dataset. Mutually exclusive with `count`.

    exact : bool
        If True, the resulting sample size will exactly match the target value.
        Exact sampling will load the whole pixel table into memory!
        If False, binomial sampling will be used instead and the sample size
        will be randomly distributed around the target value.

    map_func : function
        A map implementation.

    chunksize : int
        The number of pixels loaded and processed per step of computation.

    """
    if issubclass(type(clr), str):
        clr = cooler.Cooler(clr)

    if count is not None and frac is None:
        frac = count / clr.info["sum"]
    elif count is None and frac is not None:
        count = np.round(frac * clr.info["sum"])
    else:
        raise ValueError("Either frac or tot_count must be specified!")

    if frac >= 1.0:
        raise ValueError(
            "The number of contacts in a sample cannot exceed "
            "that in the original dataset."
        )

    if exact:
        pixels = sample_pixels_exact(clr.pixels()[:], count)
        cooler.create_cooler(out_clr_path, clr.bins()[:], pixels, ordered=True)

    else:
        pipeline = (
            cooler.tools.split(
                clr, include_bins=False, map=map_func, chunksize=chunksize
            )
            .pipe(_extract_pixel_chunk)
            .pipe(sample_pixels_approx, frac=frac)
        )

        cooler.create_cooler(out_clr_path, clr.bins()[:], iter(pipeline), ordered=True)
